# COPYRIGHT 2020 by Pietro Mandracci

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

""" PYthon tools for SImulation and CAlculus: probability density functions.

    This module contains some probability distribution functions.   

    Documentation is also available in the docstrings.
"""

#import math
from numpy import sqrt, exp, pi

from ..constants import *

#+------------------------------------------+
#| Probability distribution functions (pdf) |
#+------------------------------------------+

def pdf_uniform(x, lower, upper):
    """ Uniform probability distribution function

        f(x) = 1 / (upper-lower)

        Parameters
        ----------
 
        x:      variable
        low:    lower possible value of x
        upp:    upper possible value of x

        Returns
        -------

        pdf_uniform: value of the pdf
    """
        
    return 1.0 / ( upper - lower  )


def pdf_normal(x, mu, sigma):
    """ Normal probability distribution function

        f(x) = 1 / (sqrt(2*pi)*sigma) * exp(- (x-mu)**2 / (2*sigma**2) )

        Parameters
        ----------
 
        x:      variable
        mu:     average
        sigma:  standard deviation

        Returns
        -------

        pdf_normal: value of the pdf
    """
        
    return (1.0 / ( sqrt(2.0*pi) * sigma )
            * exp(- (x - mu)**2 / (2.0*sigma*sigma)))


def pdf_maxwell_energy(e, kt):
    """ Maxwell probability distribution function for the particle energy

        f(e) = 2/sqrt(pi) * (kt)**(3/2) * sqrt(e) * exp(-e/kt)

        Parameters
        ----------
 
        e:      energy
        kt:     temperature in energy units 
                (must be the same units used for e)

        Returns
        -------

        pdf_maxwell_energy: value of pdf
    """

    return 2.0 * sqrt( e / (pi*kt*kt*kt) ) * exp(-e/kt)


def pdf_druyvesteyn_energy_old(e, A, B):
    """ Druyvesteyn probability distribution function for the particle energy

        f(e) = A * sqrt(e) * exp(-B*e)

        Parameters
        ----------
 
        e:      energy
        A:      parameter
        B:      parameter

        Returns
        -------

        pdf_druyvesteyn_energy_old: value of pdf
    """

    return A * sqrt(e) * exp(-B*e*e)


def pdf_druyvesteyn_energy(e, B, M, E, l):
    """ Druyvesteyn probability distribution function for 
        the electron energy
        taken from: 
        FRIEDMAN A.A, KENNEDY L.A, "PLasma Physics and Engineering", 
        Taylor and Francis, 2004, pag. 184

        f(e) = B * exp(- 3*m_e/M * ( e / (q_e*E*l) )**2 )
 
        Parameters
        ----------
 
        B:      normalization coefficient
        e:      electron energy
        M:      background gas molecules mass
        E:      electric field intensity
        l:      mean free path        

        Returns
        -------

        pdf_druyvesteyn_energy: value of pdf
    """

    return (B * sqrt(e)
            * exp(- 3*ELECTRON_MASS / M * ( e / (ELECTRON_CHARGE*E*l) )**2 ))
        

def pdf_margenau_energy(e, B, M, E, w, l):
    """ Margenau probability distribution function for the electron energy
        taken from: 
        FRIEDMAN A.A, KENNEDY L.A, "PLasma Physics and Engineering", 
        Taylor and Francis, 2004, pag. 184

        f(e) = B * exp(- 3*m_e/M * ( (e**2+e*m*w**2*l**2) / (q_e*E*l)**2 )
 
        Parameters
        ----------
 
        B:      normalization coefficient
        e:      electron energy
        M:      background gas molecules mass
        E:      electric field intensity
        w:      electric field pulsation
        l:      mean free path        

        Returns
        -------

        pdf_margenau_energy: value of the pdf
    """

    return (B * sqrt(e)
            * exp(- 3*ELECTRON_MASS / M
                  * ((e**2+ELECTRON_CHARGE*ELECTRON_MASS*w*w*l*l)
                     / (ELECTRON_CHARGE*E*l)**2)))


def pdf_maxwell_speed(v, m, kt):
    """ Maxwell probability distribution function for the particle speed

        f(v) = sqrt(2/pi * (m/kt)**3) * v**2 * exp(-m*v**2 / (2*kt))

        Parameters
        ----------
 
        v:   speed
        m:   mass of the particle
        kt:  temperature in energy units 
             (must be compatible to the ones used for v and m)

        Returns
        -------

        pdf_maxwell_speed: value of the pdf
    """

    return sqrt( 2/pi * (m/kt)**3 ) * v*v * exp(- m*v*v / (2*kt) )


def pdf_maxwell_reduced(x):
    """ Maxwell probability distribution function for the adimensional particle speed

        f(x) = 4/sqrt(pi) * x**2 * exp(-x**2)

        where x = 2/sqrt(pi) v / v_mean

        Parameters
        ----------
 
        x:     adimensional speed x = 2/sqrt(pi) v / v_mean
        Returns
        -------

        pdf_maxwell_reduced: value of the pdf
    """
    
    return 4 / sqrt(pi) * x*x * exp(- x*x )


def pdf_maxwell_velocity_component(vx, m, kt):
    """ Maxwell probability distribution function for a component of the particle velocity vector

        f(vx) = (m / (2*pi*kt))**1/2 * exp(-m*v**2 / (2*kt))

        Parameters
        ----------
 
        vx:  component of the velocity vector
        m:   mass of the particle
        kt:  temperature in energy units 
             (must be compatible to the ones used for v and m)

        Returns
        -------

        pdf_maxwell_speed: value of the pdf
    """

    return sqrt(m / (2*pi*kt)) * exp(- m*vx*vx / (2*kt) )



# COPYRIGHT 2020 by Pietro Mandracci

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

""" Tools for simple statistical analysis on bivariate experimental data.

    This module contains some classes, methods and functions,useful to 
    perform some basic analysis on a bivariate set of experimental data.

    Documentation is also available in the docstrings.
"""

# +-------------------------+
# | Import required modules |
# +-------------------------+

# Mudules from the standard Python library
import math

# Modules from the Python community
import numpy
import matplotlib.pyplot as plt
#import matplotlib.patches as patches 

#from scipy.integrate   import quad
#from scipy.stats       import chi2

# Mudules from plasmapro package
from pysica.parameters import *
from pysica.managers import data_manager

# +---------------+ 
# |Bivariate data |
# +---------------+

class DataBivariate:
    """ Class defining a bivariate set of experimental data to be analyzed. """

    def __init__(self, data_matrix=None, filename='', name_x='x', name_y='y', skip=0):
        """ Defines an instance of the class DataBivariate 
            and calculates some basic statistical parameters.

            data_matrix:     a numpy 2d array: if not given, will be read from a file
            filename:        name of the file from which the data must be read, 
                             if data_matrix is not given, this must not be blank
            name_x:          a string used to identify the x values
            name_y:          a string used to identify the y values
            skip:            used only if data_matrix is not given
                             number of rows to skip while reading data from the file
            self.read_error: used for error checking
                             (status, message)
                             status: 0 = no error
                                     1 = could not open file
                                     2 = syntax error
                                     3 = non-numerical value
                                     4 = invalid number of rows/columns
                                     5 = invalid number of rows to skip
                                     6 = non-array argument was given
                                     7 = the array is not 2-d
                                     8 = wrong array structure read from file
                             message: an error message or 'OK'
        """

        status, message = 0, OK 

        if (data_matrix == None):
            matrix = data_manager.DataGrid()
            self.read_error = matrix.read_file(filename, transpose=True, skip=skip)
            if (self.read_error[0] != 0):
                status  = self.read_error[0]
                message = self.read_error[1]
                self.read_error = (status, message)                
                return
            self.data = matrix.data_array
            del matrix
            n = len(self.data)
            if (n != 2):
                status = 8
                message = 'wrong array structure in file \"'+filename+' \"'
                self.read_error = (status, message)
                return
        else:   
            try:
                n = len(data_matrix.shape)
            except AttributeError:
                status = 6
                message = 'non-array argument given'
                self.read_error = (status, message)
                return
            if (n == 2): 
                self.data = data_matrix
            else:
                status = 7
                message = 'array must be 2-dimensional'
                self.read_error = (status, message)
                return

        self.name_x =   name_x
        self.name_y =   name_y
            
        self.n_data_x = len(self.data[0])
        self.n_data_y = len(self.data[1])

        self.max_x =    self.data[0].max()        
        self.max_y =    self.data[1].max()

        self.min_x =    self.data[0].min()        
        self.min_y =    self.data[1].min()

        self.width_x =  self.max_x - self.min_x
        self.width_y =  self.max_y - self.min_y

        # Median and percentiles
        # numpy.linspace(5,100,20) = array of percentile values (5% step)
        self.percentiles_x = numpy.percentile(self.data[0],
                                              numpy.linspace(5,100,20))
        self.percentiles_y = numpy.percentile(self.data[1],
                                              numpy.linspace(5,100,20))

        
        self.median_x = self.percentiles_x[9]
        self.median_y = self.percentiles_y[9]
        
        self.mean_x =   numpy.mean(self.data[0])
        self.mean_y =   numpy.mean(self.data[1])

        # Residuals and their powers
        self.residuals_x   = self.data[0] - self.mean_x
        self.residuals_2_x = self.residuals_x * self.residuals_x
        self.residuals_3_x = self.residuals_2_x * self.residuals_x
        self.residuals_4_x = self.residuals_3_x * self.residuals_x

        # Residuals and their powers
        self.residuals_y   = self.data[1] - self.mean_y
        self.residuals_2_y = self.residuals_y * self.residuals_y
        self.residuals_3_y = self.residuals_2_y * self.residuals_y
        self.residuals_4_y = self.residuals_3_y * self.residuals_y        
                
     
        # Variance of data, with and without bias
        self.var_bias_x = self.residuals_2_x.sum() / self.n_data_x
        self.var_x      = self.var_bias_x * self.n_data_x / (self.n_data_x-1)

        self.var_bias_y = self.residuals_2_y.sum() / self.n_data_y
        self.var_y      = self.var_bias_y * self.n_data_x / (self.n_data_y-1)

        # Standard deviation, with and without bias
        self.stdv_x      = numpy.sqrt(self.var_x)
        self.stdv_y      = numpy.sqrt(self.var_y)
        self.stdv_bias_x = numpy.sqrt(self.var_bias_x)
        self.stdv_bias_y = numpy.sqrt(self.var_bias_y)

        # Standard deviation of the mean
        self.stdv_mean_x = self.stdv_x / numpy.sqrt(self.n_data_x)
        self.stdv_mean_y = self.stdv_y / numpy.sqrt(self.n_data_y)
        

    def print_stats(self):
        """ Prints some basic statistical information about data.

            Returns
            -------

            (status, message)
            status:  0 = no error
            message: a string containing an error message or 'Ok'
        """

        status, message = 0, OK

        print("Stats of data: \"" + self.name_x + "\"")
        print("Number of values:               ", self.n_data_x)
        print("Minimum value:                  ", self.min_x)
        print("Maximum value:                  ", self.max_x)
        print("Data range width:               ", self.width_x)
        print("Median:                         ", self.median_x)                
        print("5%  percentile:                 ", self.percentiles_x[0])
        print("10% percentile:                 ", self.percentiles_x[1])
        print("25% percentile:                 ", self.percentiles_x[4])
        print("75% percentile:                 ", self.percentiles_x[14])
        print("90% percentile:                 ", self.percentiles_x[17])                
        print("95% percentile:                 ", self.percentiles_x[18])
        print("Mean:                           ", self.mean_x)
        print("Standard deviation:             ", self.stdv_x)
        print("Standard deviation of the mean: ", self.stdv_mean_x)        

        print("\n")
        
        print("Stats of data: \"" + self.name_y + "\"")
        print("Number of values:               ", self.n_data_y)
        print("Minimum value:                  ", self.min_y)
        print("Maximum value:                  ", self.max_y)
        print("Data range width:               ", self.width_y)
        print("Median:                         ", self.median_y)                
        print("5%  percentile:                 ", self.percentiles_y[0])
        print("10% percentile:                 ", self.percentiles_y[1])
        print("25% percentile:                 ", self.percentiles_y[4])
        print("75% percentile:                 ", self.percentiles_y[14])
        print("90% percentile:                 ", self.percentiles_y[17])                
        print("95% percentile:                 ", self.percentiles_y[18])
        print("Mean:                           ", self.mean_y)
        print("Standard deviation:             ", self.stdv_y)
        print("Standard deviation of the mean: ", self.stdv_mean_y)                

        return (status, message)          

    def plot_scatter(self, logx=False, logy=False, legend='', title='', 
                     color = PLOT_COLOR, line = PLOT_LINE,
                     symbol = PLOT_SYMBOL):
        """ Plots a scatter plot of the data.

            xlabel          string to be used as label for the x axis
            ylabel          string to be used as label for the y axis
            legend          string to be used as legend for the data plotted
            title           string to be used as title
            logx:           set logaritmic x axis
            logy:           set logaritmic y axis
            return:         (status, message)
                            status:  0 = no error
                                     1 = unknown plot interface
                            message: a string containing an error message or 'Ok'
        """

        status, message = 0, OK         

        if (title != ''): plt.title(title)
        plt.xlabel(self.name_x)
        plt.ylabel(self.name_y)
        if (logx and logy):
            if (legend == ''):
                plt.loglog(self.data[0], self.data[1], 
                           marker=symbol, linestyle=line, color=color)
            else:
                plt.loglog(self.data[0], self.data[1], 
                           label=legend, marker=symbol, linestyle=line, color=color)
        elif logx:
            if (legend == ''):
                plt.semilogx(self.data[0], self.data[1], 
                             marker=symbol, linestyle=line, color=color)
            else:
                plt.semilogx(self.data[0], self.data[1], 
                             label=legend, marker=symbol, linestyle=line, color=color)
        elif logy:
            if (legend == ''):
                plt.semilogy(self.data[0], self.data[1], 
                             marker=symbol, linestyle=line, color=color)
            else:
                plt.semilogy(self.data[0], self.data[1], 
                             label=legend, marker=symbol, linestyle=line, color=color)
        else:
            if (legend == ''):
                plt.plot(self.data[0], self.data[1], 
                    marker=symbol, linestyle=line, color=color)
            else:
                plt.plot(self.data[0], self.data[1], 
                    label=legend, marker=symbol, linestyle=line, color=color)
        if (legend != ''): plt.legend()               
        plt.show()

        return (status, message)


    def plot_box(self, stdv=False, stdv_mean=False,
                 quartiles=True, deciles=False, maxmin=False,
                 grid=True):
        """ Plots a boxplot of the data.

            Parameters
            ----------

            stdv:      plot the mean with standard deviation
                       as errorbars
            stdv_mean: plot the mean with standard deviation of the mean 
                       as errorbars
            quartiles: plot the median with 25% and 75% as errorbars
            deciles:   plot the median with 10% and 90% as errorbars
            maxmin:    plot the median with min and max values as errorbars
            grid:      if set to True, plots a grid on the graph

            Returns
            -------

            (status, message)
            status:  0 = no error
                     1 = non numerical parameter given
                     2 = unknown plot interface
            message: a string containing an error message or 'Ok'
        """

        status, message = 0, OK

        title = 'Box plot'
        plt.title(title)
        plt.xlabel(self.name_x)
        plt.ylabel(self.name_y)
        if grid: plt.grid()
        if stdv:
            self._plot_box(plt,
                           self.mean_x, self.mean_y,
                           self.stdv_x, self.stdv_x,
                           self.stdv_y, self.stdv_y,
                           box=False,   bars=True,
                           bar_color='orange',
                           label='mean,   ' + 'stdv')
        if stdv_mean:
            self._plot_box(plt,
                           self.mean_x,      self.mean_y,
                           self.stdv_mean_x, self.stdv_mean_x,
                           self.stdv_mean_y, self.stdv_mean_y,
                           box=False,        bars=True,
                           bar_color='magenta',
                           label='mean,   ' + 'stdv_mean')     
        if maxmin:
            dx_low   = self.median_x - self.min_x
            dx_upp   = self.max_x    - self.median_x
            dy_low   = self.median_y - self.min_y
            dy_upp   = self.max_y    - self.median_y
            self._plot_box(plt,
                           self.median_x, self.median_y,
                           dx_low,        dx_upp,
                           dy_low,        dy_upp,                       
                           box=False,      bars=True,
                           bar_color='black',
                           label='median, ' + 'min-max')           
        if deciles:
            dx_low   = self.median_x          - self.percentiles_x[1]
            dx_upp   = self.percentiles_x[17] - self.median_x
            dy_low   = self.median_y          - self.percentiles_y[1]
            dy_upp   = self.percentiles_y[17] - self.median_y
            self._plot_box(plt,
                           self.median_x, self.median_y,
                           dx_low, dx_upp,
                           dy_low, dy_upp,                       
                           box=True,         bars=True,
                           line='--',
                           box_color='blue', bar_color='blue',
                           label='median, ' + '10%-90%')
        if quartiles:
            dx_low   = self.median_x          - self.percentiles_x[4]
            dx_upp   = self.percentiles_x[14] - self.median_x
            dy_low   = self.median_y          - self.percentiles_y[4]
            dy_upp   = self.percentiles_y[14] - self.median_y
            self._plot_box(plt,
                           self.median_x, self.median_y,
                           dx_low, dx_upp,
                           dy_low, dy_upp,                       
                           box=True, bars=True,
                           line='--',
                           box_color='green', bar_color='green',
                           label='median, ' + '25%-75%')
        plt.legend()
        plt.show()

        return (status, message)
    

    def _plot_box(self, plt, x, y, dx_low, dx_upp, dy_low, dy_upp,
                  box=False, bars=True,
                  box_color='black', bar_color='black', line='-',
                  label=''):
        """ Draw a data point with error bars and a rectangle

            x:         x-value of the data point
            y:         y-value of the data point
            dx_low:    lower-end error for x
            dx_upp:    upper-end error for x
            dy_low:    lower-end error for y
            dy_low:    upper-end error for y
            box:       draw a rectangle defining the error aera
            bars:      draw error bars 
            box_color: color of the rectangle
            bar_color: color of the errorbars 
            line:      style of the box line
                       {'-', '--', '-.', ':', '', (offset, on-off-seq), ...}
            label:     string to use in the legend
        """

        if box:
            box_origin = (x-dx_low, y-dy_low)
            box = plt.Rectangle(box_origin, dx_low+dx_upp, dy_low+dy_upp,
                                linewidth=1,
                                edgecolor=box_color,
                                linestyle=line,
                                facecolor='none')
            plt.gca().add_patch(box)
        if bars:
            # NOTE: errobars has a strange parameter order:
            # errorbar(x, y, yerr=None, xerr=None,...) yerr is *before* xerr       
            plt.errorbar(x, y,
                         numpy.array([[dy_low, dy_upp]]).T,
                         numpy.array([[dx_low, dx_upp]]).T,
                         marker='s', capsize=10, color=bar_color,
                         label=label)
    

    def calculate_correlation(self):
        """ Calculates the correlation coefficient of data. 

            Initialized data attributes
            ---------------------------

            self.correlation: 

            return: (status, message)
                    status:  0 = no error
                             1 = unknown plot interface
                    message: a string containing an error message or 'Ok'
        """

        status, message = 0, OK         

        return (status, message)        





# COPYRIGHT 2020 by Pietro Mandracci

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

""" PYthon tools for SImulation and CAlculus: file handling utilities

    Documentation is also available in the docstrings.
"""


# +--------------------------------+
# | Import required Python modules |
# +--------------------------------+

from os import fsync

def write_to_disk(f):
    """ Force writing to disk the content of file f """

    f.flush()
    fsync(f.fileno())



# COPYRIGHT 2020 by Pietro Mandracci

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


""" This module defines the error messages used in this package. """

OK = 'Ok'

NOERROR = (0, OK)

# Errors on plot windows
ERROR_NOT_A_PLOT      = (10, 'object is not a plot window')
ERROR_CLOSED_PLOT     = (11, 'trying to operate on a closed plot window')
ERROR_NO_PLOTS        = (12, 'no open plots')
ERROR_UNKNOWN_TERM    = (13, 'unknown terminal type')
ERROR_UNKNOWN_TYPE    = (14, 'unknown plot type')

# Errors about plotting functions or data
ERROR_NO_REPLOT       = (20, 'no functions or curves to replot')
ERROR_NO_FUNCTION     = (21, 'no functions to plot')
ERROR_NO_DATA         = (22, 'no data to plot')
ERROR_INVALID_STYLE   = (23, 'invalid plot style')
ERROR_WRONG_TYPE      = (24, 'invalid data type')
ERROR_REPLOT_VOLATILE = (25, 'replot impossible with volatile curves')
ERROR_LOGSET_VOLATILE = (26, 'logscale toggle impossible with volatile curves')

# Errors about parameters inconsistency (e.g. xmin>xmax)
ERROR_PLOT_PARAMETERS = (30, 'parameters error')

# Errors about file operations
ERROR_FILE            = (40, 'file error')








# COPYRIGHT 2020 by Pietro Mandracci

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
    Tools for the simulation of plasma discharges.

    This package contains some modules and subpackages which may be useful for the simulation of plasma discharges.

    * scripts
        - ccpla         -> script designed to simulate the motion of charged particles (electrons and ions) in 
                           a capacitivley coupled plasma discharge
        - ccpla_analysis-> analysis of output from ccpla script 
                           (ccpla must have been run with the option -s to save simulation results on ascii files)

    * modules
        - ccpla_defaults-> some constants used in the package
        - ccpla_init    -> intialization functions
        - ccpla_print   -> functions used to print simulation info in text windows
        - ccpla_gui     -> functions and classes used to run the simulaion within a GUI

    * sub-packages
        - fortran       -> modules compiled from Fortran using f2py: they provide functions to 
                           simulate the motion of electrons and ions in cold plasmas using leap-frog and PIC.
                           Functions in this package are called by the modules contained in the "discharge" subpackage;
                           normally it is not required to access them directly
        - discharge     -> modules used to define particles ensambles (charged and neutrals), reactors properties
                           and to simulate the charged particles motion
        - plot          -> modules used to plot simulation data


    Documentation is also available in the docstrings.
"""




# COPYRIGHT 2020 by Pietro Mandracci

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

""" Default values of several global variables """

from math import pi

# +-------------------------------+
# | Configuration file parameters |
# +-------------------------------+

class ConfigurationOptions:
    """ Class used to append configuration variables """

    def __init__(self):
        # Default values of parameters that can be modified by user via configuration file
        self.Nmax_particles   = 3000        # Max number of electrons allowed during simulation
        self.rescale_factor   = 10.0        # Factor for rescaling of particle numbers and weights
        self.sim_duration     = 1.0E-6      # Requested total simulation duration [s]
        self.save_delay       = 1           # Data save periodicity (0 = never, 1 = every cycle, 2 = every 2 cycles, etc..) 
        self.dt_output        = 1.0E-9      # Time between data outptut [s]
        self.dt               = 0.0         # Timestep for simulation [s] (must be lower than dt_output):
                                            #        if zero, dt is variable and calculated by the program
        self.dt_var           = True        # (boolean) Decide if a variable timestep should be used
        self.isactive_recomb  = False       # (boolean) Decide if electron/ion recombination processes must be activated
        self.lateral_loss     = False       # (boolean) Decide if electrons and ions will be lost when they reach the border
        self.maxcollfreq      = 0.01        # Maximum allowed collision frequency: used for the calculus of dt
                                            #         in variable dt mode
        self.min_scattered    = 100         # Minimum number of scattering (including null) processes
                                            #        required to apply many particles method
        self.T_neutrals       = 300         # Temperature of neutrals [K]
        self.p_neutrals       = 10          # Total pressure of gases [Pa]
        self.start_ion_deg    = 1.0E-17     # Starting ionization degree (before plasma discharge is lit)
        self.e_min_sigma      = 0.1         # Minimum energy for which cross sections for electron-neutral collisions
                                            #        will be interpolated [eV]
        self.e_max_sigma      = 1.0E3       # Maximum energy for which cross sections for electron-neutral collisions
                                            #        will be interpolated [eV]
        self.e_min_sigma_ions = 0.1         # Minimum energy for which cross sections for ion-neutral collisions
                                            #         will be interpolated [eV]
        self.e_max_sigma_ions = 1.0E2       # Maximum energy for which cross sections for ion-neutral collisions
                                            #        will be interpolated [eV]
        self.min_ion_density  = 0.0         # Minimum ion density required to activate electron/ion recombination processes
        self.N_sigma          = 1000        # Number of electron cross-section values to be calculated by interpolation
        self.N_sigma_ions     = 100         # Number of ion cross-section values to be calculated by interpolation
        self.distance         = 0.01        # Distance between electrodes [m]
        self.length           = 0.1         # Lateral length of square electrodes [m]
        self.V_bias           = 100         # Electric bias applied to electrodes [V]
        self.frequency        = 0.0         # Electric field frequency [Hz] (zero means DC)
        self.phase            = 90          # Electric field phase at t=0 [deg] (if frequency==0, this is ignored)
        self.N_cells          = 100         # Number of cells used in PIC scheme
        self.dot_points       = False       # (boolean) Decide if points must be used in plots instead of crosses
        self.n_max_points     = 1000        # Maximum number of points allowed in hisotric plots
        self.decimation_factor= 10          # When max number of points is reached, a point is removed every this number
            

# +---------------------------------+
# | Command line options parameters |
# +---------------------------------+

# Default values of parameters that can be modified via command-line options
verbosity               = 1      # How much should the program bore with text-based output
plot_xsec               = False  # Plot cross sections at start
debug_level_python      = 0      # For debugging purposes
debug_level_fortran     = 0      # 
print_only              = False  # Only print information; do not start iterations
save_defaults           = False  # Save default values to a configuration file and exit
gui_mode                = False  # Start the GUI
batch_mode              = False  # Suppress any input from user
redirect_output         = False  # Save output to a log file instead of on showing it stdout
redirect_error_messages = False  # Save python error messages to a log file instead of showing them on stderr
text_window_width       = 160    # Width of the output text window in GUI
text_window_height      = 39     # Hight of the output text window in GUI
text_window_font_size   = 12     # Size of font used in the output text window in GUI


            
# +----------------------+
# | Immutable parameters |
# +----------------------+

# Directories, filemames and separators
EXEC_PATH                 = '/usr/bin/'
EDITOR_NAME               = 'mousepad'
#EDITOR_NAME               = 'gedit'
SCRIPTNAME                = 'ccpla'                       # Name of the script
EXT                       = '.csv'                        # Extension to use for saved data files
EOL                       = '\n'                          # End-of-line char for data files
SEP                       = '\t'
SAVE_DIRECTORY_NAME       = SCRIPTNAME + '.out'           # Directory into which simulation data is saved
FILENAME_DEFAULTS         = SCRIPTNAME + '.defaults'      # Name of the file to which default parameters are saved with -s option
FILENAME_CONFIG           = SCRIPTNAME + '.conf'          # Configuration file 
FILENAME_NEUTRALS         = SCRIPTNAME + '.neutrals'      # Filename for neutrals properties
FILENAME_I                = SCRIPTNAME + '.I'             # Filename for electric current
FILENAME_V                = SCRIPTNAME + '.V'             # Filename for electric potential spatial distribution
FILENAME_OUTPUT_LOG       = SCRIPTNAME + '_output.log'    # File to which output will be saved
FILENAME_ERROR_LOG        = SCRIPTNAME + '_errors.log'    # File to which python error messages will be saved
NAME_STAT_ELE             = '_means_ele'                  # Name of output file: time evolution of several parameters
NAME_STAT_NEU             = '_means_neu'                  # Name of output file: time evolution of several parameters
NAME_DISTRIB_ELE          = '_eedf'                       # Name of output file: electrons energy distribution
NAME_DISTRIB_ION          = '_iedf'                       # First part of name of output files: ions energy distribution
NAME_ION_ELASTIC          = '_sigma4ions_Ar_elastic'
NAME_ION_CHARGE_EX        = '_sigma4ions_Ar_charge-exchange'

# Simulation parameters
NMAXPARTICLES             = 100000                        # Maximum allowed number of computational particles
MAX_RESCALE_FACTOR        = 1.0E6                         # Maximum allowed rescale factor
DEFAULT_RESCALE_FACTOR    = 10.0                          # Default scaling factor for rescaling of particles weight
DEFAULT_MIN_SCATTERED     = 100                           # Default minimum number of scattering (including null) processes
                                                          #        required to apply many particles method
START_WEIGHT              = 1.0                           # Initial value of computational weight for electrons and ions
NMAXCELLS                 = 1000                          # Maximum allowed number of cells in PIC scheme
NMINCELLS                 = 10                            # Minimum allowed number of cells in PIC scheme
MINTRATIO                 = 100                           # Minimum allowed ratio between electric field period 
                                                          #       and simulation timstep (dt)
MAX_IONIZATION_DEGREE     = 1.0E-4                        # Maximum allowed initial value of ionization degree

MAX_DISSOCIATION_TYPES    = 3                             # Maximum allowed number of dissociation channels for each neutral type
MAX_EXCITATION_TYPES      = 3                             # Maximum allowed number of excitation channels for each neutral type
MOLECULE_TYPES            = ('a', 'm', 'p')               # Types of molecules
INITIAL_VELOCITY          = 0.0                           # / m*s**-1 starting speed of electrons
INITIAL_ANGLE             = pi/6                          # / rad     starting angle between electrons velocity and z-axis direction

# GUI parameters
TEXT_WINDOW_WIDTH_MIN     = 120                           # Width of the output text window: minumum value
TEXT_WINDOW_WIDTH_MAX     = 200                           # Width of the output text window: maximum value
TEXT_WINDOW_HEIGHT_MIN    = 20                            # Height of the output text window: minumum value
TEXT_WINDOW_HEIGHT_MAX    = 80                            # Height of the output text window: maximum value
TEXT_WINDOW_FONT_TYPE     = 'mono'                        # Type of font used in the output text window
TEXT_WINDOW_FONT_SIZE_MIN = 6                             # Size of font used in the output text window
TEXT_WINDOW_FONT_SIZE_MAX = 18                            # Size of font used in the output text window
TEXT_WINDOW_FORE          = 'black'                       # Foregroun color of text in the output text window
TEXT_WINDOW_BACK          = 'light grey'                  # Background color in the output text window

FONT_MESSAGE              = ('monospace','12')
FONT_MESSAGE_SMALL        = ('monospace','10')
FONT_MESSAGE_BIG          = ('monospace','14')

IDLE_TIME                 = 1 #ms                         # Idle time betweem each main iteration cycle (in ms)
                                                          #     this is necessary to allow responsivity of the GUI
IDLE_TIME_START           = 100 #ms                       # Time between activation of main and about windows at program start
IDLE_TIME_SHUTDOWN        = 100 #ms                       # Time allowed to quit the kernele beefore killing it
IDLE_TIME_GRAPHS          = 500 #ms                       # Time allowed to reset plots before lowering them
MIN_DT_OUTPUT_EXP         = -15                           # log10 of the minimum allowed value of time between text output 
MAX_DT_OUTPUT_EXP         = -6                            # log10 of the maximum allowed value of time between text output 
MIN_PLOT_DELAY            = 1                             # Min value of number of text output cycles between each update of plots
MAX_PLOT_DELAY            = 100                           # Max value of number of text output cycles between each update of plots
RES_PLOT_DELAY            = 10                            # Minimum amount of which the value is modified in the ruler
DEF_PLOT_DELAY            = 1                             # Default value of number of text outputs between each update of plots

# Plot parameters
N_MAX_OUTPUT              = 1000                          # Maximum number of data values registered for historic plots
MAX_DECIMATION_FACTOR     = 20                            # Maximum allowed value of the decimation factor
N_BINS_LOG                = 100                           # Number of bins over which x log scale is activated in EEDF and IEDF plots
DEFAULT_TERMINAL          = 'x11'                         # Gnuplot terminal type
PERSIST                   = 0                             # Persistency of gnuplot graphs, after end of script (1=Yes, 0=No)
SCREEEN_WIDTH             = 1024                          # Default screen width
SCREEEN_HEIGHT            = 768                           # Default screen height
VERTICAL_SHIFT            = 25                            # Vertical shift toward top of screen
HBORDER                   = 12                            # Horizontal border of plot windows
VBORDER                   = 50                            # Vertical border of plot windows (including window header)
LABEL_XPOS                = 1                             # x position of the label with time value
LABEL_YPOS                = 1                             # y position of the label with time value
DEL_DATA_FILES_DELAY      = 1                             # Time to wait (in seconds) before deleting data files
                                                          # when closing cross section plots
# Debug parameters
KERNEL_DEBUG_LEV          = 1                             # Minimum debug level to get kernel debug messages



# COPYRIGHT 2020 by Pietro Mandracci

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
PYthon tools for SImulation and CAlculus: tools for data analysis

This sub-package contains some modules which can be helpful to manage 
various types of data, 
such as spectra, distribution functions, data histograms

Modules
-------

*univariate*
    analysis of univariate data 
*bivariate*
    analysis of bivariate data               
*spectra*
    manage a set of bivariate data sets and make 
    some calculations on them
"""




# COPYRIGHT 2020 by Pietro Mandracci

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
    PYthon tools for SImulation and CAlculus: statistical functions.

    This module contains some statistical functions.

    Documentation is also available in the docstrings.
"""


# +--------------------------------+
# | Import required Python modules |
# +--------------------------------+

import numpy

# +----------------------+
# | Statistics functions |
# +----------------------+
          

def calculate_bins_number(n_data, method='sqrt', width=None, stdv=None):
    """ Calculates the number of bins to create an histogram

        Parameters
        ----------

        n_data:      number of data
        method:      method used to calculate the  number of intervals
                        'sqrt':  n_bins = sqrt(n_points)
                        'log2':  n_bins = 1 + log2(n_points)
                        'root3': n_bins = n_points**(1/3)
                        'norm':  n_bins = n_data / (3.5 * stdv / n_points**(1/3) )
        with:        width of the data interval (i.e. maximum value - minimum value)
                     used only if the 'norm' method is requested
        stdv:        standard deviation of the data
                     used only if the 'norm' method is requested

        Returns
        -------
        number of intervals, on error None is returned
    """

    if   (method == 'sqrt'):  return int( numpy.sqrt(n_data) )
    elif (method == 'log2'):  return int( 1 + numpy.log2(n_data) )
    elif (method == 'root3'): return int(2.0 * pow(n_data, 1.0/3.0))
    elif (method == 'norm'):
        if ( (width is None) or (stdv is None) ): return None
        else:
            delta  = 3.5 * stdv / pow(n_data, 1.0/3.0)
            return int(width / delta)
    else: return None



# COPYRIGHT 2020 by Pietro Mandracci

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
PYSICA - PYthon tools for SImulation and CAlculus
=================================================

Modules in the main package
---------------------------

*constants*
    some physical constant sused in various modules and packages
*parameters*
    some parameters used in various modules and packages
    
Sub-packages in the main package
-------------------------------

*functions*
    some general purpose functions
*fortran*
    modules compiled from Fortran using f2py
*managers*
    modules which can be used to manage the input/output of data from/to ascii files
    and to print physical quantities managing the unit prefixes
*analysis*
    modules which can be helpful to manage distribution functions and data histograms
*plasmapro*
    package of tools for simulatio of plasma discharges.

Documentation is also available in the docstrings.
"""




# COPYRIGHT 2020 by Pietro Mandracci

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

""" PYthon tools for SImulation and CAlculus: random generation utilities.

    This module contains some functions to generate random number with specific distributions.

    Documentation is also available in the docstrings.
"""

#import math
import numpy
from random import random

from ..constants import *
from .pdf import *

def random_maxwell_velocity(v_mean, modulus=False):
    """ Generate the x,y,z components of a velocity vector following Maxwell distribution

        Parameters
        ----------

        v_mean:   mean value of the velocity module
        modulus:  if True, the modulus of the velocity vector is returned also

        Returns
        -------

        if modulus==False: (vx, vy, vz):   components of the random velocity vector
        if modulus==False: (vx, vy, vz,v): components and modulus of the random velocity vector
    """

    v = random_maxwell_speed(v_mean)

    costheta = 2 * random() - 1
    sintheta = numpy.sqrt(1 - costheta * costheta)
    phi      = 2 * numpy.pi * random()
    vx       = v * sintheta * numpy.cos(phi)
    vy       = v * sintheta * numpy.sin(phi)
    vz       = v * costheta

    if modulus: return (vx, vy, vz, v)
    else:       return (vx, vy, vz)      

    
def random_maxwell_speed(v_mean):
    """ Generate a random speed according to the Maxwell distribution.

        Parameters
        ----------
        
        v_mean: mean value of the velocity module       

        Returns
        -------
        
        random speed value
    """
    
    xmax = 4.0
    ymax = 4.0 / (numpy.sqrt(numpy.pi) * numpy.e)
    f = 0
    y = 1
    while(f < y):
        x = random() * xmax
        y = random() * ymax
        f = pdf_maxwell_reduced(x)

    return sqrt(pi) / 2 * v_mean * x



# COPYRIGHT 2020 by Pietro Mandracci

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

""" Read and show data saved by ccpla script """

# +--------------------------------+
# | Import required Python modules |
# +--------------------------------+

# Mudules from the standard Python library
import math

# Modules provided by the Python community
import numpy
import matplotlib.pyplot as plt
import Gnuplot

# Import required constants and parameters
from pysica.parameters import *
from pysica.constants import *
from pysica.plasmapro.ccpla_defaults import *
#from pysica.plasmapro.plot.plot_parameters import *

# Modules provided by plasmapro package
from pysica.managers import data_manager
from pysica.managers.unit_manager import print_unit, print_exp
from pysica.analysis import univariate
from pysica.functions.pdf import pdf_maxwell_energy
from pysica.plasmapro.discharge.reactors import CcpProperties
from pysica.plasmapro.discharge.target_particles import TargetParticles
from pysica.plasmapro.ccpla_init import *

plt.rcParams["mathtext.default"] = 'regular'
plt.rcParams["font.size"]        = 20


class CcplaSavedData:
    """Class of data saved by ccpla program"""

    def __init__(self, name=None, verbose=False, read_edf=False, read_potential=False):
        """Initialize the collection of data to be analyzed"""

        # Reading EEDF and IEDF as well as potential distribution can be suppressed to save memory and execution time
        self.read_edf = read_edf
        self.read_V   = read_potential

        # Define some constants
        self.DEBYE_CONST_A = 1.51E13                                # (J m)**(-1/2)
        self.DEBYE_CONST_B = 4/3 * numpy.pi * self.DEBYE_CONST_A**3 # (J m)**(-3/2)
        self.DEBYE_CONST_C = 56.41                                  # m**(3/2) s**-1         

        # Read data from configuration files
        self.parameters   = ConfigurationOptions()
        (status, message) = initialize_parameters(self.parameters, verbose=verbose, saveonly=False,
                                                  filename_config=FILENAME_CONFIG)
        if (status != 0):
            self.error = (status, message)
            return

        if verbose: print('\nLoading reactor properties\n')
        self.ccp = CcpProperties(self.parameters.distance, self.parameters.length,
                                 self.parameters.V_bias,   self.parameters.frequency, self.parameters.phase,
                                 self.parameters.N_cells,  self.parameters.lateral_loss)

        if verbose: print('Reading neutral properties from file \"'+ FILENAME_NEUTRALS + '\"\n')
        self.neutrals = TargetParticles(self.parameters.N_sigma, self.parameters.N_sigma_ions,
                                        self.parameters.T_neutrals, self.parameters.p_neutrals,
                                        self.parameters.min_scattered, self.parameters.isactive_recomb,
                                        filename=FILENAME_NEUTRALS)
        (status, message) = self.neutrals.read_error
        if (status != 0):
                self.error = (status, message)
                return
        (status, message) = self.neutrals.read_properties(FILENAME_NEUTRALS, '\t',
                                                          self.parameters.e_min_sigma, self.parameters.e_max_sigma,
                                                          self.parameters.e_min_sigma_ions, self.parameters.e_max_sigma_ions,
                                                          debug=False)
        if (status != 0):
                self.error = (status, message)
                return

        # Read data saved during simulation
        generate_save_dir_name(self.parameters, abs_path=False)
        generate_save_file_names(self.parameters)

        self.parameters.filename_stat_ele +=  EXT
        self.parameters.filename_stat_neu +=  EXT
        self.parameters.filename_I        +=  EXT
        if self.read_edf:
            self.parameters.filename_distrib_ele +=  EXT
        if self.read_V:
            self.parameters.filename_V +=  EXT                        

        if verbose: print('Reading electron mean data from file \''+ self.parameters.filename_stat_ele +'\'\n')
        self.means_ele = data_manager.DataGrid()
        (status, message) = self.means_ele.read_file(self.parameters.filename_stat_ele, sep=SEP, transpose=True, skip=1)
        if (status != 0):
            string = 'Error reading file \"' + self.parameters.filename_stat_ele + '\": '+message
            self.error =  (status, string)
            return

        if verbose: print('Reading electron neutrals data from file \''+ self.parameters.filename_stat_neu +'\'\n')
        self.means_neu = data_manager.DataGrid()
        (status, message) = self.means_neu.read_file(self.parameters.filename_stat_neu, sep=SEP, transpose=True, skip=1)
        if (status != 0):
            string = 'Error reading file \"' + self.parameters.filename_stat_neu + '\": '+message
            self.error =  (status, string)
            return

        if verbose: print('Reading electric current data from file \''+ self.parameters.filename_I +'\'\n')
        self.means_I = data_manager.DataGrid()
        (status, message) = self.means_I.read_file(self.parameters.filename_I, sep=SEP, transpose=True, skip=1)
        if (status != 0):
            string = 'Error reading file \"' + self.parameters.filename_I + '\": '+message
            self.error =  (status, string)
            return   

        if self.read_edf:
            if verbose: print('Reading EEDF data from file \''+ self.parameters.filename_distrib_ele +'\'\n')
            self.eedf = data_manager.DataGrid()
            (status, message) = self.eedf.read_file(self.parameters.filename_distrib_ele, sep=SEP, pad_value=numpy.nan)
            if (status != 0):
                string = 'Error reading file \"' + self.parameters.filename_distrib_ele + '\": '+message
                self.error =  (status, string)
                return
            self.iedf = []
            for i in range(self.neutrals.types):
                filename = self.parameters.filename_distrib_ion + '_' + self.neutrals.names[i] + '+' + EXT
                if verbose: print('Reading IEDF data from file \''+ filename +'\'\n')
                #iedf = data_manager.DataGrid()
                self.iedf.append( data_manager.DataGrid() )
                (status, message) = self.iedf[i].read_file(filename, sep=SEP, pad_value=numpy.nan)
                if (status != 0):
                    string = 'Error reading file \"' + filename + '\": '+message
                    self.error =  (status, string)
                    return

        if self.read_V:
            if verbose: print('Reading electric potential data from file \''+ self.parameters.filename_V +'\'\n')
            self.V = data_manager.DataGrid()
            (status, message) = self.V.read_file(self.parameters.filename_V, sep=SEP)
            if (status != 0):
                string = 'Error reading file \"' + self.parameters.filename_V + '\": '+message
                self.error =  (status, string)
                return
                

        self.n_rows           = len(self.means_ele.data_array[0])
        self.sim_duration     = self.means_ele.data_array[0, self.n_rows-1] * 1E-9
        self.output_timestep  = self.means_ele.data_array[0, self.n_rows-1] * 1E-9 / self.n_rows
        self.n_cols           = len(self.means_neu.data_array)
        self.mol_types        = int((self.n_cols - 1) / 2)

        # lambda_D = A * sqrt(<E_e>**3 / n_e)
        # NOTE: ELECTRON_CHARGE is negative, multiplying by it is necessary to convert energy from eV to J
        self.debye_length     = (  self.DEBYE_CONST_A
                                 * numpy.sqrt(- ELECTRON_CHARGE
                                              * self.means_ele.data_array[3]
                                              / self.means_ele.data_array[14] )
                                )
        # N_D = 4/3 * pi * lamnda_D**3 * n_e
        self.debye_number     = 4/3 * numpy.pi * self.debye_length**3 * self.means_ele.data_array[14]
        # f_pla ~ sqrt(n_e)
        self.plasma_frequency = self.DEBYE_CONST_C * numpy.sqrt(self.means_ele.data_array[14])

        if verbose: print('Data loading completed\n')                

        self.error = (status, message)

    def print_parameters(self):
        print('')
        print('Electrodes distance                 : ' + print_unit(self.ccp.distance,'m'))
        print('Electrodes lateral length           : ' + print_unit(self.ccp.length,'m'))
        print('Plasma volume                       : ' + str(self.ccp.volume) + ' m**3')
        print('Number of PIC cells                 : ' + str(self.parameters.N_cells))
        print('Cell dimension                      : ' + print_unit(self.ccp.delta_grid, 'm', 4 ))
        print('Electric bias                       : ' + print_unit(self.ccp.V_peak,'V'))
        print('Electric bias frequency             : ' + print_unit(self.ccp.frequency,'Hz'))
        print('Electric bias phase (at t=0)        : ' + str(self.ccp.phase))
        print('')
        print('Gas temperature                     : ' + str(self.neutrals.temperature) + ' K')
        print('Total pressure                      : ' + str(self.neutrals.total_pressure) + ' Pa')
        print('Number of tabulated e- xsec values  : ' + str(self.neutrals.n_sigma))
        print('Number of tabulated ion xsec values : ' + str(self.neutrals.n_sigma_ions))
        print('')
        print('Maximum number of electrons         : ' + str(self.parameters.Nmax_particles))
        print('Required starting ionization degree : ' + print_exp(self.parameters.start_ion_deg,4))
        print('')
        print('Timestep [0=automatic]              : ' + print_unit(self.parameters.dt, 's'))
        print('Mean time between data acquisitions : ' + print_unit(self.output_timestep, 's'))
        print('Number of data acquisitions         : ' + str(self.n_rows-1))
        print('Overall simulated time              : ' + print_unit(self.sim_duration, 's'))        

    def get_row(self, time):
        for row in range(self.n_rows):
            if (self.means_ele.data_array[0][row] >= time): break
        return row
    
    # +----------------------------------------+
    # | Plasma parameters time evolution plots |
    # +----------------------------------------+
        
    def plot_electron_number(self, real=True, computational=True, line='None', symbol='.',
                             color_real='red', color_comp='blue'):
        plt.ioff()
        plt.title('Number of electrons')
        plt.xlabel('Time / ns')
        plt.ylabel('Number of electrons')
        if computational:
            plt.semilogy(self.means_ele.data_array[0], self.means_ele.data_array[1], 
                         marker=symbol, linestyle=line, color=color_comp, label=r'Computational $e^-$')
        if real:
            plt.semilogy(self.means_ele.data_array[0], self.means_ele.data_array[1] * self.means_ele.data_array[2], 
                         marker=symbol, linestyle=line, color=color_real, label=r'Real $e^-$')
        plt.legend()
        plt.grid()
        plt.show()

    def plot_electron_weight(self,line='None', symbol='.', color='red'):
        plt.ioff()
        plt.title('Electron weight')
        plt.xlabel('Time / ns')
        plt.ylabel('Weight')
        plt.semilogy(self.means_ele.data_array[0], self.means_ele.data_array[2], 
                     marker=symbol, linestyle=line, color=color, label = r'Weight $e^-$')
        plt.legend()
        plt.grid()
        plt.show()     

    def plot_electron_density(self, line='None', symbol='.', color='red'):
        plt.title('Electron density')
        plt.xlabel('Time / ns')
        plt.ylabel(r'$n_e$ / $m^{-3}$')
        plt.semilogy(self.means_ele.data_array[0], self.means_ele.data_array[14], 
                     marker=symbol, linestyle=line, color=color, label = r'$n_e$')
        plt.grid()
        plt.legend()
        plt.show()

    def plot_plasma_frequency(self,line='None', symbol='.', color='red'):
        plt.title('Plasma frequency')
        plt.xlabel('Time / ns')
        plt.ylabel(r'$\nu_P$ / $s^{-1}$')
        plt.semilogy(self.means_ele.data_array[0], self.plasma_frequency, 
                     marker=symbol, linestyle=line, color=color, label = r'$\nu_P$')
        plt.grid()
        plt.legend()
        plt.show()        

    def plot_debye_length(self,line='None', symbol='.', color='red'):
        plt.title('Debye length')
        plt.xlabel('Time / ns')
        plt.ylabel(r'$\lambda_D$ / m')
        plt.semilogy(self.means_ele.data_array[0], self.debye_length, 
                     marker=symbol, linestyle=line, color=color, label = r'$\lambda_D$')
        plt.grid()
        plt.legend()
        plt.show()
        
    def plot_debye_number(self,line='None', symbol='.', color='red'):
        plt.title('Debye number')
        plt.xlabel('Time / ns')
        plt.ylabel(r'$N_D$')
        plt.semilogy(self.means_ele.data_array[0], self.debye_number, 
                     marker=symbol, linestyle=line, color=color, label = r'$N_D$')
        plt.grid()
        plt.legend()
        plt.show()    
               
    def plot_electron_mean_energy(self, plot_sigma=False, plot_min=False, plot_max=False, semilog=False,
                                  line='None', symbol='.', color='red', ecolor='orange',
                                  maxcolor='cyan', mincolor='blue'):
        plt.ioff()
        plt.title('Electron mean energy')
        plt.xlabel('Time / ns')
        plt.ylabel(r'<$E_e$> / eV')
        if semilog:
                plt.semilogy(self.means_ele.data_array[0], self.means_ele.data_array[3],
                             marker=symbol, linestyle=line, color=color, label=r'<$E_e$>')
        else:
            if plot_sigma:
                plt.errorbar(self.means_ele.data_array[0], self.means_ele.data_array[3], self.means_ele.data_array[4], 
                             marker=symbol, linestyle=line, color=color, ecolor=ecolor, label = r'<$E_e$>')
            else:
                plt.plot(self.means_ele.data_array[0], self.means_ele.data_array[3], 
                         marker=symbol, linestyle=line, color=color, label = r'<$E_e$>')

            if plot_min: plt.plot(self.means_ele.data_array[0], self.means_ele.data_array[5], 
                                  marker=symbol, linestyle=line, color=mincolor, label = r'$E_e^{min}$')
            if plot_max: plt.plot(self.means_ele.data_array[0], self.means_ele.data_array[6], 
                                  marker=symbol, linestyle=line, color=maxcolor, label = r'$E_e^{max}$')
        plt.legend()
        plt.grid()
        plt.show()

        
    def plot_electron_angle(self, plot_sigma=True, plot_min=True, plot_max=True,
                            line='None', symbol='.', color='red', ecolor='orange',
                            maxcolor='cyan', mincolor='blue'):
        plt.ioff()
        plt.title('Mean angle between electron velocity and electric field')
        plt.xlabel('Time / ns')
        plt.ylabel(r'<$\theta$> / °')
        if plot_sigma:                        
            plt.errorbar(self.means_ele.data_array[0], self.means_ele.data_array[7], self.means_ele.data_array[8], 
                         marker=symbol, linestyle=line, color=color, ecolor=ecolor, label=r'<$\theta$>')
        else:
            plt.plot(self.means_ele.data_array[0], self.means_ele.data_array[7], 
                     marker=symbol, linestyle=line, color=color, label = r'<$\theta$>')

        if plot_min: plt.plot(self.means_ele.data_array[0], self.means_ele.data_array[9], 
                              marker=symbol, linestyle=line, color=mincolor, label = r'$\theta_{min}$')
        if plot_max: plt.plot(self.means_ele.data_array[0], self.means_ele.data_array[10], 
                              marker=symbol, linestyle=line, color=maxcolor, label = r'$\theta_{max}$')
        plt.legend()
        plt.grid()                
        plt.show()

        
    def plot_tau(self, line='None', symbol='.', color='red'):
        plt.ioff()
        plt.title('Timestep and time between collisions')
        plt.xlabel('Time / ns')
        plt.ylabel('Time / fs')
        plt.semilogy(self.means_ele.data_array[0], self.means_ele.data_array[11], 
                     marker=symbol, linestyle=line, color=color, label = 'dt')
        plt.semilogy(self.means_ele.data_array[0], self.means_ele.data_array[12], 
                     marker=symbol, linestyle=line, color=color,  label = r'$\tau$')
        plt.legend()
        plt.grid()
        plt.show()

        
    def plot_collision_frequency(self, line='None', symbol='.', color='red'):
        plt.ioff()
        plt.title('Collision frequency')
        plt.xlabel('Time / ns')
        plt.ylabel('f %')
        plt.plot(self.means_ele.data_array[0], self.means_ele.data_array[13], 
                 marker=symbol, linestyle=line, color=color, label = 'f')
        plt.legend()
        plt.grid()
        plt.show()
        
    # +---------------------------+
    # | Electric properties plots |
    # +---------------------------+
    
    def plot_current_density(self, absolute=False, line='None', symbol='.', color='red'):
        plt.ioff()
        plt.title('Current density')
        plt.xlabel('Time / ns')
        plt.ylabel(r'j / $A m^{-2}$')
        if absolute:
            plt.plot(self.means_I.data_array[0] / (self.ccp.length * self.ccp.length),
                     numpy.abs(self.means_I.data_array[1]), 
                     marker=symbol, linestyle=line, color=color, label = 'j')            

        else:
            plt.plot(self.means_I.data_array[0] / (self.ccp.length * self.ccp.length),
                     self.means_I.data_array[1], 
                     marker=symbol, linestyle=line, color=color, label = 'j')
        plt.legend()
        plt.grid()
        plt.show()


    def plot_potential(self,time, line='-', symbol='.', color='red'):
        row = self.get_row(time)
        plt.ioff()
        plt.title('Electric potential')
        plt.xlabel('z / mm')
        plt.ylabel(r'$\Delta V$ / V')
        plt.plot(self.ccp.grid_points*1.0E3, self.V.data_array[row], 
                 marker=symbol, linestyle=line, color=color)
        plt.grid()
        plt.show()           
        
    # +---------------------------------------+
    # | Dissociation rate/rate constant plots |
    # +---------------------------------------+
        
    def plot_dissocation_rates(self, line='None', symbol='.', color='red'):
        plt.title('Dissociation rates')
        plt.xlabel('Time / ns')
        plt.ylabel('R / m**3 s**-1')
        i_name = 0
        for i in range(self.mol_types):
            col = 1 + i*2
            # search the name of the molecule
            for j in range(i_name, self.neutrals.types):
                if (self.neutrals.molecule_type[j] != 'a'):
                    i_name = j
                    break
            name = self.neutrals.names[i_name]
            i_name += 1
            plt.semilogy(self.means_neu.data_array[0], self.means_neu.data_array[col],
                         marker=symbol, linestyle=line, color=color, label = name)
        plt.grid()
        plt.legend()
        plt.show()

        
    def plot_dissocation_rate_const(self, line='None', symbol='.', color='red'):
        plt.title('Dissociation rate constants')
        plt.xlabel('Time / ns')
        plt.ylabel('k / m**3 s**-1')
        i_name = 0                
        for i in range(self.mol_types):
            col = 1 + i*2 +1
            # search the name of the molecule
            for j in range(i_name, self.neutrals.types):
                if (self.neutrals.molecule_type[j] != 'a'):
                    i_name = j
                    break
            name = self.neutrals.names[i_name]
            i_name += 1                        
            plt.semilogy(self.means_neu.data_array[0],self.means_neu.data_array[col],
                         marker=symbol, linestyle=line, color=color, label = name)
        plt.grid()
        plt.legend()
        plt.show()

    # +---------------------------+
    # | Energy distribution plots |
    # +---------------------------+
                
    def plot_ensamble(self, index, time=None, row=None, log=False, color='red', symbol=','):
        if not self.read_edf:
            print('Distribution functions are not available')
            return
        if (time is None):
            if (row is None):
                print('Give either simulation time or data row number !')
                return
        else:
            row = self.get_row(time)
        time = self.means_ele.data_array[0][row]                
        string = 'row #' + str(row) + "; time = " + str(time) + ' ns'
        if (index == 0):
            name = 'e- '
            ensamble = self.eedf.data_array[row]
        else:
            name     = self.neutrals.names[index-1] + '+ '
            ensamble = self.iedf[index-1].data_array[row]
        plt.ioff()
        plt.title('Ensamble of ' + name)
        plt.xlabel('Particle index')
        plt.ylabel('Energy / eV')
        if log:
            plt.semilogy(ensamble, marker=symbol, linestyle='None', color=color, label=name+string)
        else:    
            plt.plot(ensamble, marker=symbol, linestyle='None', color=color, label=name+string)
        plt.legend()
        plt.grid()
        plt.show()
                
                
    def plot_eedf(self, time=None, row=None, pdf='Maxwell', intervals=0, int_method='sqrt', method='fixed', plot_interface='pylab'):
        """ Compare the particle distribution with a distribution fuction. 

            A histogram will be created, the number of bins can be given or automatically calculated

            Parameters
            ----------

            time:                   simulation time at which the eedf/iedf is requested
            row:                    data row containint the eedf/iedf requested
                                    must be given if time is not
            pdf:                    probability distribution function describing the model
                                   'Maxwell' -> Maxwell pdf for particle kinetic energy

            intervals:              maximum allowed number of intervals (0 means no limit given)
            int_method:             method used to calculate the maximum number of intervals, used only if intervals=0
                                    'sqrt':  n_bins = sqrt(n_points)
                                    'log2':  n_bins = 1 + log2(n_points)
                                    'root3': n_bins = n_points**(1/3)
                                    'norm':  n_bins = n_data / (3.5 * stdv / n_points**(1/3) )
            method:                 method used to calculate the intervals
                                    'fixed': a fixed number of equiparted intervals is used
                                    'tails': width of each interval is a multiple of self.width/sqrt(self.n_data)
                                    expected frequency for each interval (except the last one) 
                                    is not less than 5                      
                                    'iterate': all intervals have the same width
                                    none of the intervals has zero frequency 
                                    not more than 20% of the intervals has a frequency lower than 5
            plot_interface:         graphic interface to be used to plot the histogram
                                    'pylab'   -> use matplotlib
                                    'gnuplot' -> use gnuplot

        """
        
        if not self.read_edf:
            print('Distribution functions are not available')
            return
        
        if (time is None):
                if (row is None):
                        print('Give either simulation time or data row number !')
                        return
        else:
                row = self.get_row(time) 
        print('\nRow # ' + str(row) + ' of ' + str(self.n_rows-1))
        self.n_electrons = int(self.means_ele.data_array[1, row])
        self.h= univariate.DataSet(self.eedf.data_array[row][0:self.n_electrons])
        if (pdf == 'Maxwell'):
                self.kt = 2.0 * self.h.mean / 3.0
                self.pdf = lambda x: pdf_maxwell_energy(x, self.kt)
        else:
                print('Unknown distribution type !')
                return
        print(self.h.expected_frequencies(self.pdf, intervals=intervals, int_method=int_method, method=method))
        print(self.h.observed_frequencies())
        print(self.h.chisquare_estimation())
        print('Time        = ' + print_unit(self.means_ele.data_array[0, row]*1E-9,'s'))
        print('Electrons   = ' + str(self.n_electrons))
        print('Weight      = ' + str(self.means_ele.data_array[2, row]))
        print('Mean energy = ' + print_unit(self.means_ele.data_array[3, row],  'eV', 3))
        print('Stdv energy = ' + print_unit(self.means_ele.data_array[4, row],  'eV', 3))
        print('Min energy  = ' + print_unit(self.means_ele.data_array[5, row],  'eV', 3))
        print('Max energy  = ' + print_unit(self.means_ele.data_array[6, row],  'eV', 3))
        print('Mean angle  = ' + print_unit(self.means_ele.data_array[7, row],  'deg', 3))
        print('Stdv angle  = ' + print_unit(self.means_ele.data_array[8, row],  'deg', 3))
        print('Min angle   = ' + print_unit(self.means_ele.data_array[9, row],  'deg', 3))
        print('Max angle   = ' + print_unit(self.means_ele.data_array[10, row],         'deg', 3))
        print('dt          = ' + print_unit(self.means_ele.data_array[11,row]*1E-15, 's', 3))
        print('tau         = ' + print_unit(self.means_ele.data_array[12,row]*1E-15, 's', 3))
        print('p           = ' + str(self.means_ele.data_array[13,row]) + ' %')
#       print 'E_mean      = ' + print_unit(h.mean,             'eV', 3)
#       print 'E_max       = ' + print_unit(h.max,              'eV', 3)
#       print 'E_min       = ' + print_unit(h.min,              'eV', 3)
        print('k*T         = ' + print_unit(self.kt,            'eV', 3))
        print('N           = ' + str(self.h.n_data))
        print('DF          = ' + str(self.h.freedom_degrees))
        print('Chisquare   = ' + str(self.h.chisquare))
        print('P-value     = ' + str(self.h.p_value))
        print(self.h.plot_histogram(interface=plot_interface))
        



#!/usr/bin/python

# COPYRIGHT 2020 by Pietro Mandracci

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


""" This script is intended to test the functions module.

    If the package has been correctly installed, you should be able to run 
    this script from the system terminal by typing::

    $ python -m gnuplot_manager.test

    or inside the python console by typing::

    >>> from gnuplot_manager.test import main
    >>> main()
"""

import numpy
from .functions import *

def wait():
    return input('Press RETURN to continue...\n')

def main():

    x = numpy.linspace(0, 100, 101)
    a = numpy.linspace(0, 2.7, 271)

    print('\n**Test of gnuplot_manager functions**\n')
    wait()

    print('* -> Function new_plot')
    p2d = new_plot(xpos=480, ypos=100,
                   plot_type='2D', title='TEST 2D PLOT WINDOW', redirect_output=False)
    p3d = new_plot(xpos=500, ypos=120,
                   plot_type='3D', title='TEST 3D PLOT WINDOW', redirect_output=False)
    wait()
   
    print('* -> Function plot2d')
    plot2d(p2d, x, x*x, 'parabola')
    wait()

    print('* -> Function plot2d with replot')
    plot2d(p2d, x, x*x*x/100, 'x^3/100', replot=True)
    wait()

    print('* -> Function plot3d')
    plot3d(p3d, numpy.cos(4*a), numpy.sin(4*a), a, 'elicoidal (I)')
    wait()

    print('* -> Function plot3d with replot')
    plot3d(p3d, numpy.sin(4*a), numpy.cos(4*a), a, 'elicoidal (II)', replot=True)
    wait()
    
    print('* -> Function plot_curves (2D)')
    list1 = [ [x, x*x, 'parabola', 'with points'] , [x, x*x*x/100, 'x^3/100', 'with lines']  ]
    plot_curves(p2d, list1)
    wait()

    print('* -> Function plot_curves (3D)')
    list2 = [ [ numpy.cos(4*a), numpy.sin(4*a), a, 'elicoidal (I)', 'with points'] ,
              [ numpy.sin(4*a), numpy.cos(4*a), a, 'elicoidal (II)','with lines' ]  ]
    plot_curves(p3d, list2)
    wait()

    print('* -> Function plot_function (2d)')
    plot_function(p2d, 'x**2', None)
    wait()

    print('* -> Function plot_function (2d) with replot')
    plot_function(p2d, 'x**3/100', None, replot=True)
    wait()  

    print('* -> Function plot_function (3d)')
    plot_function(p3d, 'x**2+y**2', None)
    wait()

    print('* -> Function plot_function (3d) with replot')
    plot_function(p3d, '(x**3+y**3)/10', None, replot=True)
    wait()

    print('* -> Function plot_functions (2d)')
    list1 = [ ['x**2', None, 'with points'] , ['x**3/10', None, 'with lines']  ]
    plot_functions(p2d, list1)
    wait()

    print('* -> Function plot_functions (3d)')
    list2 = [ ['x**2+y**2', None, 'with points'] , ['(x**3+y**3)/10', None, 'with lines']  ]
    plot_functions(p3d, list2)
    wait()    
    
    print('* -> Function plot_set')
    plot_set(p3d, xmin=1, xmax=10, ymin=1, ymax=1000, logz=True, replot=True)
    wait()

    print('* -> Function plot_label')
    plot_label(p2d, x=20, y=20, label='THIS IS A LABEL', replot=True)
    wait()

    print('* -> Function plot_label (erase=True)')
    plot_label(p2d, label=None, erase=True, replot=True)
    wait()

    print('* -> Function plot_list')
    plot_list()
    wait()

    print('* -> Function plot_list with expanded option')
    plot_list(expanded=True)
    wait()

    print('* -> Function plot_close_all')
    plot_close_all()
    wait()

    print('**End of test**\n')
    
if __name__ == '__main__':
    main()



# COPYRIGHT 2020 by Pietro Mandracci

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

""" Classes for the simulation of charged particles in a plasma discharge.

    This module contains a class to manage charged particles (electrons and ions) moving inside a plasma discharge.
    It provides methods to extract some characteristics of their ensables, such as mean velocity or kinetic energy,
    etc...

    Documentation is also available in the docstrings.
"""


# +--------------------------------+
# | Import required Python modules |
# +--------------------------------+

import math
import numpy
from pysica.parameters import *
from pysica.constants import *
from pysica.plasmapro.ccpla_defaults import *
from pysica.io.io_files import write_to_disk


# +------------------+
# | Moving particles |
# +------------------+

class MovingParticles:
        """ A class defining a collection of particles moving in a cold plasma"""
        
        def __init__(self, types, Nmax_particles, start_weight=START_WEIGHT, rescale_factor=DEFAULT_RESCALE_FACTOR):
                """Initialises the properties of a collection of particles moving in a cold plasma .

                   The number of particle types must be given (e.g. 3 if electrons and 2 types of ions are considered)
                   The masses and charges of all particle types are set to the electrons values.
                   All the particles are set as inactive (non existing).
                   The starting position is set to zero.

                        Parameters
                        ----------

                        types:                  number of particle types
                        Nmax_particles:         maximum number of particles which will be possible to manage (for each type)
                                                (dimension of the array in which particles data will be stored)
                        start_weight:           initial value of the particles computational weight 
                        fdt:                    requested value of f*dt in variable dt mode
                        rescale_factor:         the factor by which the number of particles will be divided and the weight 
                                                       will be multiplied if they exceed the maximum allowed number
                                                       by default it is the same for all particle types, 
                                                       but you can set a different value
                                                       for each type by manually changing the values of the 
                                                       self.rescale_factor array



                        Initialized data attributes
                        ---------------------------

                        self.types:             number of particles types
                        self.names:             array containg particle names (may be used for some output)
                        self.n:                 dimension of the array in which particles data are stored
                        self.charge:            particle charge / C (initialized to the electron charge)
                        self.mass:              particle mass / kg  (initialized to the electron mass)
                        self.cm_ratio:          particle charge / mass ratio / C*kg**-1
                        self.number_density:    number density of this particle type / m**-3
                        self.weight:            weight of the particle for simulation purposes,
                                                        i.e. the number of real particles represented by the simulation particle
                        self.rescale_factor:    factor by which the number of particles (of a given type) 
                                                        will be divided and their weight will be multiplied
                                                        if their number exceeds the maximum allowed
                        self.active:            an array identifying which particles are existing at a given time
                                                        True  = active (existing) particle
                                                        False = inactive (non existing) particle
                        self.restart_lf         an array identifying which particles must restart leap-frog scheme
                                                        at next iteration
                                                        True  = restart at next iteration
                                                        False = do not restart

                        self.z                          position along direction parallel to electric field / m
                        self.r                          position along direction normal to electric field / m
                        self.v:                         velocity module / m*s**-1
                        self.vz:                        z components of velocities / m*s**-1
                        self.vx:                        x components of velocities / m*s**-1
                        self.vy:                        y components of velocities / m*s**-1
                        self.vr:                        r components of velocities (x-y plane) / m*s**-1
                        self.theta:                     angle between velocity direction and z axis / rad
                        self.dt:                        timestep for particles motion
#                        self.fdt:                       maximum value of electron collision freq * dt
#                                                            used to calculate dt in variable dt mode
                        self.time:                      actual value of simulation time
                        self.tau_mis:                   approximated value of time between electron collisions
                        self.p_coll:                    approximated value of elctron collision probability

                """
               
                self.types              = int(types)
                self.n                  = int(Nmax_particles)
                self.names              = numpy.zeros(self.types, 'U8')
                self.charge             = numpy.ones(self.types, 'd') * ELECTRON_CHARGE
                self.mass               = numpy.ones(self.types, 'd') * ELECTRON_MASS
                self.cm_ratio           = self.charge / self.mass
                self.number_density     = numpy.zeros(self.types, 'd')
                self.debye_length       = numpy.zeros(self.types, 'd')
                self.plasma_frequency   = numpy.zeros(self.types, 'd')
                self.weight             = numpy.ones(self.types, 'd') * start_weight
                self.rescale_factor     = numpy.ones(self.types, 'd') * rescale_factor
                self.active             = numpy.zeros( (self.types, Nmax_particles), 'bool', order='F')
                self.restart_lf         = numpy.ones ( (self.types, Nmax_particles), 'bool', order='F')
                self.z                  = numpy.zeros( (self.types, Nmax_particles), 'd',    order='F')
                self.x                  = numpy.zeros( (self.types, Nmax_particles), 'd',    order='F')
                self.y                  = numpy.zeros( (self.types, Nmax_particles), 'd',    order='F')
                self.v                  = numpy.zeros( (self.types, Nmax_particles), 'd',    order='F')
                self.vz                 = numpy.zeros( (self.types, Nmax_particles), 'd',    order='F')#self.v*numpy.cos(self.theta)
                self.vx                 = numpy.zeros( (self.types, Nmax_particles), 'd',    order='F')
                self.vy                 = numpy.zeros( (self.types, Nmax_particles), 'd',    order='F')
                self.vr                 = numpy.zeros( (self.types, Nmax_particles), 'd',    order='F')
                self.theta              = numpy.zeros( (self.types, Nmax_particles), 'd',    order='F')
                self.dt                 = 0.0
                self.time               = 0.0
                self.tau_mis            = 0.0
                self.p_coll             = 0.0

                for i in range(self.types):
                        self.names[i] = 'Q_'+str(i)


        def reset(self, start_weight=START_WEIGHT, rescale_factor=DEFAULT_RESCALE_FACTOR):
                """Reset ensamble properties (only mutable ones) to initial values"""
                
                self.number_density.fill(0)
                self.debye_length.fill(0)
                self.plasma_frequency.fill(0)
                self.weight.fill(start_weight)
                self.rescale_factor.fill(rescale_factor)
                self.active.fill(False)
                self.restart_lf.fill(True)
                self.z.fill(0)
                self.x.fill(0)
                self.y.fill(0)
                self.v.fill(0)
                self.vz.fill(0)
                self.vx.fill(0)
                self.vy.fill(0)
                self.vr.fill(0)
                self.theta.fill(0)
                self.dt                 = 0.0
                self.time               = 0.0
                self.tau_mis            = 0.0
                self.p_coll             = 0.0

                        
        def n_active(self, i_type):
                """Returns the number of active particles of the type i_type.

                        Parameters
                        ----------

                        i_type:         index that refers to to the type of particle

                        self:           record of particles data
                        self.active:    an array identifying which particles are existing
                                                True  = active particle
                                                False = inactive (non existing) particle

                        Returns
                        -------
                        
                        number of active particles of type i_type
                """

                return self.active[i_type].sum()


        def v_max(self, i_type):
                """Calculates the maximum of active particles speeds.

                        Parameters
                        ----------

                        i_type:         index that refers to to the type of particle

                        self:           record of particles data
                        self.v:         array of velocity modules / m*s**-1
                        self.active:    an array identifying which particles are existing
                                                True  = active particle
                                                False = inactive (non existing) particle

                        Returns
                        -------
                        
                        maximum speed of active particles of type i_type / m*s**-1
                """

                if self.active[i_type].any():
                        return self.v[i_type][self.active[i_type]][numpy.argmax( self.v[i_type][self.active[i_type]] )]
                else:
                        return 0.0
                


        def v_min(self, i_type):
                """Calculates the minimum of active particles speeds.

                        Parameters
                        ----------

                        i_type:         index that refers to to the type of particle

                        self:           record of particles data
                        self.v:         array of velocity modules / m*s**-1
                        self.active:    an array identifying which particles are existing
                                                True  = active particle
                                                False = inactive (non existing) particle

                        Returns
                        -------

                        return: minimum speed of active particles of type i_type  / m*s**-1
                """

                if self.active[i_type].any():
                        return self.v[i_type][self.active[i_type]][numpy.argmin( self.v[i_type][self.active[i_type]] )]
                else:
                        return 0.0


        def v_average(self, i_type):
                """Calculates the average of active particles speeds.

                        Parameters
                        ----------

                        i_type:         index that refers to to the type of particle

                        self:           record of particles data
                        
                        self.v:         array of velocity modules / m*s**-1
                        self.active:    an array identifying which particles are existing
                                                True  = active particle
                                                False = inactive (non existing) particle

                        Returns
                        -------

                        average speed of active particles of type i_type  / m*s**-1

                """
                if self.active[i_type].any():
                        return numpy.average( self.v[i_type][self.active[i_type]] )
                else:
                        return 0.0


        def v2_average(self, i_type):
                """Calculates the average of the square of active particles velocities.

                        Parameters
                        ----------

                        i_type:         index that refers to to the type of particle

                        self:           record of particles data
                        
                        self.v:         array of velocity modules / m*s**-1
                        self.active:    an array identifying which particles are existing
                                                True  = active particle
                                                False = inactive (non existing) particle

                        Returns
                        -------

                        average of squared speeds of active particles of type i_type  / (m*s**-1)**2
                """
                
                if self.active[i_type].any():
                        return numpy.average( ( self.v[i_type][self.active[i_type]] )**2 )
                else:
                        return 0.0


        def v_sigma(self, i_type):
                """Calculates the standard deviation of active particles speeds.

                        Parameters
                        ----------      

                        i_type:         index that refers to to the type of particle

                        self:           record of particles data
                        
                        self.v:         array of velocity modules / m*s**-1
                        self.active:    an array identifying which particles are existing
                                                True  = active particle
                                                False = inactive (non existing) particle
                        Returns
                        -------
                        
                        standard deviation of velocity modules of active particles of type i_type  / m*s**-1
                """

                if (self.n_active(i_type) > 1):
                        return numpy.sqrt( sum( ( self.v[i_type][self.active[i_type]] - self.v_average(i_type) )**2 ) / \
                                           (self.n_active(i_type)-1) )
                else:
                        return 0.0


        def all_energies(self, i_type):
                """Returns an array with particles kinetic energies, including inactive ones.

                        Parameters
                        ----------

                        i_type: index that refers to to the type of particle

                        self:   record of particles data

                        self.v: array of velocity modules / m*s**-1

                        Returns
                        -------

                        array of kinetic energies of (active and inactive) particles of type i_type  / eV
                """

                if (self.charge[i_type] != 0):
                        return self.mass[i_type] / (2*numpy.abs(self.charge[i_type])) * self.v[i_type]**2                         
                else:
                        return numpy.zeros(self.n)


                
        def energies(self, i_type):
                """Returns an array with active particles kinetic energies.

                        Parameters
                        ----------

                        i_type:      index that refers to to the type of particle

                        self:        record of particles data

                        self.v:      array of velocity modules / m*s**-1

                        self.active: an array identifying which particles are existing
                                     True  = active particle
                                     False = inactive (non existing) particle

                        Returns
                        -------

                        array of kinetic energies of active particles of type i_type  / eV
                """

                if (self.charge[i_type] != 0):
                        return self.mass[i_type] / (2*numpy.abs(self.charge[i_type])) * self.v[i_type][self.active[i_type]]**2 
                else:
                        return numpy.zeros(self.n_active(i_type))



        def e_average(self, i_type):
                """Calculates the average of active particles kinetic energies.

                        Parameters
                        ----------

                        i_type:  index that refers to to the type of particle

                        self:    record of particles data

                        Returns
                        -------

                        average kinetic energy of active particles of type i_type  / eV
                """

                if ( (self.charge[i_type] != 0) and self.active[i_type].any() ):
                        return self.mass[i_type] / (2*numpy.abs(self.charge[i_type])) * self.v2_average(i_type)                         
                else:
                        return 0.0                        

        def kT(self, i_type):
                """Calculates the peak value (for Boltzmann distribution) of active particles kinetic energies.

                        Parameters
                        ----------

                        i_type:  index that refers to to the type of particle

                        self:    record of particles data

                        Returns
                        -------

                        k*T / eV
                """

                if (self.charge[i_type] != 0):
                        return 2.0 / 3.0 * self.e_average(i_type) # <E> = 3/2 kT                        
                else:
                        return 0.0
                

        def e_min(self, i_type):
                """Calculates the minimum of active particles kinetic energies.

                        Parameters
                        ----------

                        i_type:  index that refers to to the type of particle

                        self:    record of particles data

                        Returns
                        -------

                        minimum kinetic energy of active particles of type i_type  / eV
                """
                
                if ( (self.charge[i_type] != 0) and self.active[i_type].any() ):
                        return self.mass[i_type] / (2*numpy.abs(self.charge[i_type])) * self.v_min(i_type)**2                        
                else:
                        return 0.0



        def e_max(self, i_type):
                """Calculates the maximum of active particles kinetic energies.

                        Parameters
                        ----------

                        i_type:  index that refers to to the type of particle

                        self:    record of particles data

                        Returns
                        -------

                        maximum kinetic energy of active particles of type i_type  / eV
                """
                
                if ( (self.charge[i_type] != 0) and self.active[i_type].any() ):
                        return self.mass[i_type] / (2*numpy.abs(self.charge[i_type])) * self.v_max(i_type)**2                        
                else:
                        return 0.0


        def e_sigma(self, i_type):
                """Calculates the standard deviation of active particles kinetic energies.

                        Parameters
                        ----------

                        i_type:  index that refers to to the type of particle

                        self:    record of particles data

                        Returns
                        -------

                        standard deviation of kinetic energies of active particles of type i_type  / eV
                """

                if (self.n_active(i_type) > 1): # In order to have a finite standard deviation, need at least 2 particles
                        return numpy.sqrt( sum( (self.energies(i_type) - self.e_average(i_type) )**2 ) / (self.n_active(i_type)-1) )
                else:
                        return 0.0



        def theta_average(self, i_type):
                """Calculates the average of active particles angles.

                        Parameters
                        ----------

                        i_type:         index that refers to to the type of particle
        
                        self:           record of particles data
                        
                        self.theta:     angle between velocity direction and z-axis / rad
                        self.active:    an array identifying which particles are existing
                                                True  = active particle
                                                False = inactive (non existing) particle

                        Returns
                        -------

                        average angle of active particles of type i_type  / deg
                """
                
                if self.active[i_type].any():
                        return numpy.average(self.theta[i_type][self.active[i_type]])
                else:
                        return 0.0


        def theta_sigma(self, i_type):
                """Calculates the standard deviation of active particles angles.

                        Parameters
                        ----------

                        i_type:         index that refers to to the type of particle
        
                        self:           record of particles data
                        
                        self.theta:     angle between velocity direction and z-axis / rad
                        self.active:    an array identifying which particles are existing
                                                True  = active particle
                                                False = inactive (non existing) particle

                        Returns
                        -------

                        standard deviation of angles of active particles of type i_type  / rad
                """

                if (self.n_active(i_type) > 1):
                        return numpy.sqrt( sum( (self.theta[i_type][self.active[i_type]] - self.theta_average(i_type) )**2 ) / \
                                (self.n_active(i_type)-1.0) )                       
                else:
                        return 0.0
                

        def theta_max(self, i_type):
                """Calculates the maximum of active particles angles.

                        Parameters
                        ----------

                        i_type:         index that refers to to the type of particle

                        self:           record of particles data
                        self.theta:     array of angles
                        self.active:    an array identifying which particles are existing
                                                True  = active particle
                                                False = inactive (non existing) particle

                        Returns
                        -------
                        
                        maximum angle / rad
                """
                if self.active[i_type].any():
                        return self.theta[i_type][self.active[i_type]][numpy.argmax(self.theta[i_type][self.active[i_type]])]
                else:
                        return 0.0


        def theta_min(self, i_type):
                """Calculates the minimum of active particles angles.

                        Parameters
                        ----------

                        i_type:         index that refers to to the type of particle

                        self:           record of particles data
                        self.theta:     array of angles
                        self.active:    an array identifying which particles are existing
                                                True  = active particle
                                                False = inactive (non existing) particle

                        Returns
                        -------
                        
                        minimum angle of active particles of type i_type  / rad
                """

                if self.active[i_type].any():
                        return self.theta[i_type][self.active[i_type]][numpy.argmin(self.theta[i_type][self.active[i_type]])]
                else:
                        return 0.0


        def initialize_savefiles(self,
                                 filename_stat, filename_distrib_ele, filename_distrib_ion,
                                 append=False, sep=SEP, ext='.csv'):
                """Initializes the data files to which simulation data about moving particles will be saved

                   Parameters
                   ----------
                   filename_stat:         name of the file to which the simulation quantities will be saved:
                                          electron enenergy, electron number,...
                   filename_distrib_ele:  name of the file to which the electron energy distrobution function
                                          will be saved

                   filename_distrib_ion:  root used to form the names of the files to which the ion energy 
                                          distribution functions will be saved.
                   append:                if True, open the file for appendig new data to the end;
                                          the header will not be re-written
                   sep:                   character or sequence to separate data columns
                   ext:                   extension to give to filenames (deafult is '.csv')


                   Initialized data attributes
                   ---------------------------

                   self.sep:                   character or sequence to separate data columns
                   self.f_stat_name:           name of the file to which values of several quantities will be saved
                   self.f_distrib_ele_name:    name of the file to which eedf will be saved
                   self.f_distrib_ion_names:   list of the names of the files to which iedf of each ion type will be saved
                """
                self.sep         = sep
                self.f_stat_name = str(filename_stat) + ext
                
                # Open data file where to save time evolution of mean values
                if  append:
                        data_file = open(self.f_stat_name,'a')
                else:
                        data_file = open(self.f_stat_name,'w')                        
                        # Write header
                        data_file.write('\"t[ns]\"'         + self.sep )
                        data_file.write('\"N electrons\"'   + self.sep ) # number of active electrons
                        data_file.write('\"w\"'             + self.sep ) # weight
                        data_file.write('\"En[eV]\"'        + self.sep )
                        data_file.write('\"sEn[eV]\"'       + self.sep )
                        data_file.write('\"MinEn[eV]\"'     + self.sep )
                        data_file.write('\"MaxEn[eV]\"'     + self.sep )
                        data_file.write('\"theta[deg]\"'    + self.sep )
                        data_file.write('\"stheta[deg]\"'   + self.sep )
                        data_file.write('\"Mintheta[deg]\"' + self.sep )
                        data_file.write('\"Maxtheta[deg]\"' + self.sep )
                        data_file.write('\"dt[fs]\"'        + self.sep )
                        data_file.write('\"tau[fs]\"'       + self.sep )
                        data_file.write('\"coll_prob[%]\"'  + self.sep )
                        data_file.write('\"n_e[m**-3]\"'    + self.sep )
                        data_file.write(EOL)                        
                data_file.close()

                self.f_distrib_ele_name = str(filename_distrib_ele) + (ext)
                # Open data file where to save evolution of eedf
                if (append):
                        data_file = open(self.f_distrib_ele_name,'a')
                else:
                        data_file = open(self.f_distrib_ele_name,'w')
                data_file.close()                        
                        
                # Open data files where to save evolution of iedf for each ion type
                self.f_distrib_ion_names = []
                for i in range(1, self.types):
                        filename = str(filename_distrib_ion) + '_' + self.names[i] + ext
                        self.f_distrib_ion_names.append(filename)                        
                        if append:                                
                                data_file = open(filename,'a')
                        else:
                                data_file = open(filename,'w')
                        data_file.close()


        def save_data_to_files(self):
                """Saves actual data values to files """

                data_file = open(self.f_stat_name, 'a')                             
                # Save data on mean values 
                data_file.write( str( 1E9*self.time                      ) + self.sep )
                data_file.write( str( self.n_active(0)                   ) + self.sep ) # number of active electrons
                data_file.write( str( self.weight[0]                     ) + self.sep ) # weight
                data_file.write( str( self.e_average(0)                  ) + self.sep )
                data_file.write( str( self.e_sigma(0)                    ) + self.sep )
                data_file.write( str( self.e_min(0)                      ) + self.sep )
                data_file.write( str( self.e_max(0)                      ) + self.sep )
                data_file.write( str( math.degrees(self.theta_average(0))) + self.sep )       
                data_file.write( str( math.degrees(self.theta_sigma(0))  ) + self.sep )
                data_file.write( str( math.degrees(self.theta_min(0))    ) + self.sep )
                data_file.write( str( math.degrees(self.theta_max(0))    ) + self.sep )
                data_file.write( str( 1.0E15*self.dt                     ) + self.sep )
                data_file.write( str( 1.0E15*self.tau_mis                ) + self.sep )
                data_file.write( str( self.p_coll                        ) + self.sep )
                data_file.write( str( self.number_density[0]             ) + self.sep )                 
                data_file.write(EOL)
                data_file.close()

                # Save actual eedf
                data_file = open(self.f_distrib_ele_name,'a')
                self.e_distr = self.energies(0)
                for j in range(self.n_active(0)):
                        data_file.write( str(self.e_distr[j]) + self.sep )
                data_file.write( EOL )
                data_file.close()

                # Save actual iedf for all ion types
                for i in range(1,self.types):
                        i_file = i-1
                        data_file = open(self.f_distrib_ion_names[i_file], 'a')
                        self.e_distr = self.energies(i)
                        for j in range(self.n_active(i)):
                                data_file.write( str(self.e_distr[j]) + self.sep )
                        data_file.write( EOL )
                        data_file.close()




# COPYRIGHT 2020 by Pietro Mandracci

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
    PYthon tools for SImulation and CAlculus: physical functions 
    
    This module contains some functions used to calculate or to convert physical quantities.

    Documentation is also available in the docstrings.
"""


# +--------------------------------+
# | Import required Python modules |
# +--------------------------------+

import numpy

from ..constants import *


# +--------------------+
# | Physical functions |
# +--------------------+

def mean_speed_maxwell(temperature, mass):
    """ Calculates the mean speed of the molecules in a gas at thermal equilibrium 

        Parameters
        ----------

        temperature:    gas temperature / K
        mass:           molecular mass / atomic units

        Returns
        -------

        mean speed / m s**-1

    """

    return numpy.sqrt(8 * K_BOLTZMANN * temperature / (numpy.pi * mass * ATOMIC_UNIT_MASS))


def number_density(pressure, temperature):
    """ Calculates the number density of a gas at given pressure and temperature.

        Parameters
        ----------

        pressure:       gas pressure / Pa
        temperature:    gas temperature / K

        Returns
        -------

        number density: numer density of gas molecules / m**-3
    """

    return pressure / (K_BOLTZMANN * temperature)


def pressure_conversion(pressure_in, unit_in, unit_out):
    """ Converts a pressure measure between different units.

        Parameters
        ----------
                
        pressure_in:    pressure value
        unit_in:        units in which the input pressure is expressed 
                        ('Pa', 'mbar', 'mtorr', 'torr', 'atm', 'psi')
        unit_out:       units in which the output pressure must be expressed 
                        ('Pa', 'mbar', 'mtorr', 'torr')

        Returns
        -------

        pressure_out:   pressure value expressed in converted units
                        if the string given as input or output does not match any of the 
                        pressure units, the value -1 will be returned
    """
                
    if (unit_out=='mtorr'):
        if   (unit_in=='Pa'):    return pressure_in * ATM_PRESSURE_TORR * 1000 / ATM_PRESSURE_PA
        elif (unit_in=='mbar'):  return pressure_in * ATM_PRESSURE_TORR * 1000 / ATM_PRESSURE_PA * 100
        elif (unit_in=='mtorr'): return pressure_in
        elif (unit_in=='torr'):  return pressure_in * 1000
        elif (unit_in=='atm'):   return pressure_in * ATM_PRESSURE_TORR * 1000
        elif (unit_in=='psi'):   return pressure_in * ATM_PRESSURE_TORR * 1000 / ATM_PRESSURE_PSI
        else:                    return -1 


    elif (unit_out=='torr'):
        if   (unit_in=='Pa'):    return pressure_in * ATM_PRESSURE_TORR / ATM_PRESSURE_PA
        elif (unit_in=='mbar'):  return pressure_in * ATM_PRESSURE_TORR / ATM_PRESSURE_PA * 100
        elif (unit_in=='mtorr'): return pressure_in / 1000.0
        elif (unit_in=='torr'):  return pressure_in 
        elif (unit_in=='atm'):   return pressure_in * ATM_PRESSURE_TORR 
        elif (unit_in=='psi'):   return pressure_in * ATM_PRESSURE_TORR / ATM_PRESSURE_PSI
        else:                    return -1        

    elif (unit_out=='Pa'):
        if   (unit_in=='Pa'):    return pressure_in
        elif (unit_in=='mbar'):  return pressure_in * 100.0
        elif (unit_in=='mtorr'): return pressure_in * ATM_PRESSURE_PA   / (ATM_PRESSURE_TORR * 1000)
        elif (unit_in=='torr'):  return pressure_in * ATM_PRESSURE_PA / ATM_PRESSURE_TORR
        elif (unit_in=='atm'):   return pressure_in * ATM_PRESSURE_PA
        elif (unit_in=='psi'):   return pressure_in * ATM_PRESSURE_PA / ATM_PRESSURE_PSI
        else:                    return -1

    elif (unit_out=='mbar'):
        if   (unit_in=='Pa'):    return pressure_in / 100.0
        elif (unit_in=='mbar'):  return pressure_in
        elif (unit_in=='mtorr'): return (pressure_in * ATM_PRESSURE_PA / 100
                                         / (ATM_PRESSURE_TORR * 1000))
        elif (unit_in=='torr'):  return pressure_in * ATM_PRESSURE_PA / 100 / ATM_PRESSURE_TORR
        elif (unit_in=='atm'):   return pressure_in * ATM_PRESSURE_PA / 100
        elif (unit_in=='psi'):   return pressure_in * ATM_PRESSURE_PA / 100 / ATM_PRESSURE_PSI
        else:                    return -1

    elif (unit_out=='atm'):
        if   (unit_in=='Pa'):    return pressure_in * 1.0 / ATM_PRESSURE_PA
        elif (unit_in=='mbar'):  return pressure_in * 1.0 / ATM_PRESSURE_PA * 100
        elif (unit_in=='mtorr'): return pressure_in * 1.0 / ATM_PRESSURE_TORR / 1000
        elif (unit_in=='torr'):  return pressure_in * 1.0 / ATM_PRESSURE_TORR
        elif (unit_in=='atm'):   return pressure_in
        elif (unit_in=='psi'):   return pressure_in * 1.0 / ATM_PRESSURE_PSI 
        else:                    return -1 

    elif (unit_out=='psi'):
        if   (unit_in=='Pa'):    return pressure_in * ATM_PRESSURE_PSI / ATM_PRESSURE_PA
        elif (unit_in=='mbar'):  return pressure_in * ATM_PRESSURE_PSI / ATM_PRESSURE_PA * 100
        elif (unit_in=='mtorr'): return pressure_in * ATM_PRESSURE_PSI / (ATM_PRESSURE_TORR * 1000)
        elif (unit_in=='torr'):  return pressure_in * ATM_PRESSURE_PSI / ATM_PRESSURE_TORR
        elif (unit_in=='atm'):   return pressure_in * ATM_PRESSURE_PSI
        elif (unit_in=='psi'):   return pressure_in
        else:                    return -1 

    else:
        return -1




# COPYRIGHT 2020 by Pietro Mandracci

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

""" PYthon tools for SImulation and CAlculus: mathematical functions.

    This module contains some mathematical functions used to compute averages, derivatives and so on.

    Documentation is also available in the docstrings.
"""


# +--------------------------------+
# | Import required Python modules |
# +--------------------------------+

import numpy
from scipy.stats import norm


# +------------------------+
# | Mathematical functions |
# +------------------------+

def interpolate(x, x0, x1, y0, y1):
    """ Returns the linear interpolation of the given values

        Parameters
        ----------

        x0: lower value of independent variable
        x1: upper value of indepenfent variable
        y0: lower value of dependent variable
        y1: upper value of dependent variable

        Returns
        -------

        y = y0 + (y1-y0) * (x-x0) / (x1-x0)
    """

    return y0 + (y1-y0) * (x-x0) / (x1-x0)


def interpolate_3p(x, x0, x1, x2, y0, y1, y2):
    """ Returns the interpolation of the given values using 3 points

        Parameters
        ----------

        x0: lower  value of independent variable
        x1: medium value of independent variable
        x2: upper  value of independent variable
        y0: lower  value of dependent variable
        y1: medium value of dependent variable
        y2: upper  value of dependent variable

        Returns
        -------

        y = interpolated value, using 3 points
    """

    d0  = x  - x0
    d1  = x  - x1
    d2  = x  - x2
    d01 = x0 - x1
    d02 = x0 - x2
    d12 = x1 - x2
    d10 = -d01
    d20 = -d02
    d21 = -d12
    L0  =  (d1 * d2) / (d01 * d02)
    L1  =  (d0 * d2) / (d10 * d12)
    L2  =  (d0 * d1) / (d20 * d21)
    y   = L0 * y0 + L1 * y1 + L2 * y2

    return y


def _derivative_5p(dx, y0minus2, y0minus1, y0, y0plus1, y0plus2):
    """ Calculates the 5-points derivative of function f(x) at point x0.

        Parameters
        ----------

        dx         finite difference of x-value
        y0minus2   f(x0-2*dx)
        y0minus1   f(x0-dx) 
        y0         f(x0)
        y0plus1    f(x0+dx)
        y0plus2    f(x0+2*dx)

        Returns
        -------

        y1 = f'(x0)
    """

    return (y0minus2 - 8 * y0minus1 + 8 * y0plus1 - y0plus2 ) / (12 * dx)


def derivative_5p(dx, y):
    """ Calculates the derivative of a function y=f(x), the y values are given as a numpy array.

        Parameters
        ----------
        dx: distance between the x-points for which the y values were given
        y: numpy array contaning

        Returns
        -------

        y1: a numpy array containing the values of f'(x) 
            the first 2 and last 2 values are nan
    """

    n = len(y)
    y1 = numpy.ones(n) * numpy.nan
    for i in range(2,n-2):
        y1[i] = _derivative_5p(dx, y[i-2], y[i-1], y[i], y[i+1], y[i+2])
                
    return y1


def median_filter(a, window, pad=False, spikes=False, k=1):
    """ Returns the moving average of an array, using a window of 2w+1 points

        Parameters
        ----------

        a:       the array to be filtered

        window:  width of the mediana window, in points
                 the i-th point of the averaged array will be the median
                 of points from i-window to i+window (included) of a
                 this value must be lower than len(a)/2
        pad:     if False, the first and last w points of the new spectrum will be padded with nan
                 if True, the window used to calculate median is padded with the first/last point
        spikes:  if True a procedure to remove spike is applied,
                 otherwise, the standard median filter is applied 
        k:       used only if spikes is True: points will be cosidered spikes if their y value 
                 differs (in modulus) from the median of the 2w+1 windows 
                 of more than k times the maximum 


        Returns
        -------

        a_median:  the medinan filtered array, since the filter uses a number of points equal 
                   to 'window' before and after each one for which it is calculated,
                   it is not possibile to calculate it for a number of values equal to 'window'
                   at the beginning and end of the array: these will be padded with nan 
    """

    n = len(a)

    if (window < 0): window = - window
    if (window >= n / 2): window = n / 2 - 1
    window = int(window)
    if (k < 0): k = - k
    a_median = numpy.ones(n) * numpy.nan

    # This array will be used in the unspike mode to remove the central point of the window
    # in order to calculate the maximum deviation from the median
    if spikes:
        window_indexes         = numpy.ones(2*window+1,'bool')
        window_indexes[window] = False        

    for i in range(n):
        # Define the window of points from which the median
        # will be calculated, from i-window to i+window, both included
        # if the point is too near the the edges, and pad is True,
        # the first or last point is used to fill the window
        if (i < window):
            if pad:
                a_window = numpy.zeros(2*window+1,'d')
                for j in range(window-i):            a_window[j] = a[0]
                for j in range(window-i,2*window+1): a_window[j] = a[j-window+i]
            else:
                continue
        elif (i >= n - window):
            if pad:
                a_window = numpy.zeros(2*window+1,'d')
                for j in range(window+n-i):            a_window[j] = a[j-window+i]
                for j in range(window+n-i,2*window+1): a_window[j] = a[n-1]
            else:
                continue
        else:
            a_window = a[i-window:i+window+1]
            
        median = a_window[a_window.argsort()][window]
                
        if spikes:
            # Calculate the maximum deviation from the median 
            max_dev = abs(a_window[window_indexes] - median ).max()
            # Substitute the point with the median,
            # if it differs from the median more than k times
            # the maximum deviation from the median of the other points
            if (abs(a[i]-median) > k * max_dev):
                a_median[i] = median
            else:
                a_median[i] = a[i]
        else:
            a_median[i] = median

    return a_median


def moving_average(a, window):
    """ Returns the moving average of an array, using a window of 2w+1 points

        Parameters
        ----------

        a: the array to be averaged

        window: width of the average window, in points
                the i-th point of the averaged array will be the average 
                of points from i-window to i+window (included) of a
                this value must be lower than len(a)/2

        Returns
        -------

        a_averaged: the averaged array, since the average uses a number of points equal 
                    to 'window' before and after each one for which it is calculated,
                    it is not possibile to calculate it for a number of values equal to 'window'
                    at the beginning and end of the array: these will be padded with nan 
    """

    n = len(a)

    if (window < 0): window = - window
    if (window >= n / 2): window = n / 2 - 1
    window = int(window)

    a_smooth = numpy.ones(n) * numpy.nan

    # First point (i = window)
    mov_average = 0
    for j in range(0, 2 * window + 1):
        mov_average = mov_average + a[j]
    mov_average = mov_average / (2 * window + 1)
    a_smooth[window] = mov_average

    # Following points (i = window + 1 .. n - window - 1)
    for i in range(window + 1, n - window):
        add_value = (a[i+window] - a[i-window-1]) / (2 * window + 1)
        mov_average = mov_average + add_value
        a_smooth[i] = mov_average

    return a_smooth


def moving_average_skip(a, window, skip=0):
    """ Returns the moving average of an array, using a window of 2w+1 points, 
        allowing to skip some points.

        Parameters
        ----------

        a: the array to be averaged

        window: width of the average window, in points
                the i-th point of the averaged array will be the average 
                of points from i-window to i+window (included) of a
                this value must be lower than len(a)/2
        skip:   number of values to skip at the beginning and end of the array

        Returns
        -------

        a_averaged: the averaged array, since the average uses a number of points equal 
                    to 'window' before and after each one for which it is calculated,
                    it is not possibile to calculate it for a number of values equal to 'window'
                    at the beginning and end of the array: these will be padded with nan 
    """

    n = len(a)

    if (skip < 0):      skip = - skip
    if (skip >= n / 2): skip = n / 2 - 1
    skip = int(skip)

    m = n - 2 * skip
    if (window < 0):      window = - window
    if (window >= m / 2): window = m / 2 - 1
    window = int(window)

    a_smooth = numpy.ones(n) * numpy.nan

    # First point (i = window + skip)
    mov_average = 0
    for j in range(skip, 2 * window + 1 + skip):
        mov_average = mov_average + a[j]
    mov_average = mov_average / (2 * window + 1)
    a_smooth[window+skip] = mov_average

    # Following points (i = window + 1 + skip .. n - window - 1 - skip)
    for i in range(window + 1 + skip, n - window - skip):
        add_value = (a[i+window] - a[i-window-1]) / (2 * window + 1)
        mov_average = mov_average + add_value
        a_smooth[i] = mov_average

    return a_smooth


def calculate_weights(n, method='linear', sigma=None, k=3):
    """ Returns an arrays of weights to be used for the calculation of a weighted moving average.

        Parameters
        ----------

        n:      number of weight
        method: method used to calculate weights: 'uniform', 'linear' or 'gaussian'
                'uniform'   -> uniform weights
                'linear'    -> linear weights
                'gaussian'  -> gaussian weights, with stadard deviation equal to (n-1)/k
        sigma:  used in the 'gaussian' mode only
                it is the standard deviation of the normal distribution used for weights calculation
                if not given, it will be calculated
        k:      used in the 'gaussian' mode, if the sigma parameter is not given
                the standard deviation of the normal distribution used 
                for wieghts calculation will be 
                    sigma = (n-1)/k
                    default is k=3, so that the the sum of the weights is ~> 0.99

        Returns
        -------

        weights: an array containing the weights 

    """

    n = int(n)
    weights = numpy.ones(n,'d') * numpy.nan

    w = n - 1
    if (method == 'uniform'):                
        for i in range(n):
            weights[i] = 1.0 / (2*n - 1)               
    elif (method == 'linear'):
        for i in range(n):
            weights[i] = 1.0 * (w - i + 1) / ( (w + 1) * (w + 1) )
    elif (method =='gaussian'):
        # If sigma was not given, we calculate it in such a way that k*sigma = w+0.5
        # since each weight is calculated as the gaussian probability from (i-0.5) to (i+0.5)
        if (sigma is None): sigma = 1.0 * (w+0.5) / k 
        for i in range(n):
            inf = i - 0.5
            sup = i + 0.5
            P_inf = norm.cdf(inf, loc=0, scale=sigma)
            P_sup = norm.cdf(sup, loc=0, scale=sigma)
            weights[i] = P_sup - P_inf
    return weights
        
def moving_average_weighted(a, w, normalize=False, skip=0):
    """ Returns the weighted moving average of an array, using an array of weights

        Each point of the averaged array will be given by
        average[i] = sum( j, from i-l to i+l of array[j] * weights[abs(j-i)] )
               
        The weights w[i] must satisfy these properties
        w[j] = w[-j]
        sum(i, from -l to l, weights[i]) = 1

        The array given contains the values weights[0] to weights[l]


        Parameters
        ----------

        a:         the array to be averaged
        w:         the array of weights, these must be real numers belonging to the interval [0,1]
        normalize: if True, each weight is divided by the sum of the weights, so that their 
                   sum becomes 1
        skip:      numebr of points to skip at the beginning and end of the signal

        Returns
        -------

        a_averaged: the averaged array, since the average uses a number of points equal 
                    to len(w)-1 before and after each one for which it is calculated,
                    it is not possibile to calculate it for a number of values equal to 'window'
                    at the beginning and end of the array: these will be padded with nan 
    """

    n =      len(a)
    window = len(w) - 1

    # If required, normalize the weights.
    if (normalize):
        N = 2 * w.sum() - w[0]
        w = w / N

    a_smooth = numpy.ones(n,'d') * numpy.nan

    # If the array length is not enough to calculate the average
    # for at least one point, return an array of NaN
    if (n < 2 * window + 1): return a_smooth

    # From a[window] to a[n-window-1]
    for i in range(window + skip, n - window - skip):
        mov_average = 0.0
        # weighted average of a[i-window] to a[i+window]
        for j in range(i-window, i+window+1):
            mov_average = mov_average + a[j] * w[abs(j-i)]
        a_smooth[i] = mov_average

    return a_smooth


def roughness(z):
    """ Returns the roughness of a surface profile.

        Parameters
        ----------

        z: surface profile

        Returns
        -------

        z_mean, Ra, Rq, Rsk: mean height, Ra value, Rq value, Rsk (skewness), Rku (kurtosis)
        in case of error, None is returned
    """

    # Remove NaNs and infs from the signal
    z_real      = z[numpy.isfinite(z)]
    n_real      = len(z_real)
    z_mean      = z_real.mean()
    z_delta     = z_real - z_mean
    z_delta2    = z_delta  * z_delta
    z_delta3    = z_delta2 * z_delta
    z_delta4    = z_delta3 * z_delta
    Ra          = numpy.abs(z_delta).sum() / n_real
    Rq          = numpy.sqrt( z_delta2.sum() / n_real )
    Rsk         = z_delta3.sum() / (n_real * Rq * Rq * Rq)
    Rku         = z_delta4.sum() / (n_real * Rq * Rq * Rq * Rq)         

    return (z_mean, Ra, Rq, Rsk, Rku)



# COPYRIGHT 2020 by Pietro Mandracci

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
    PYthon tools for SImulation and CAlculus: io utilities

    Documentation is also available in the docstrings.
"""


# +--------------------------------+
# | Import required Python modules |
# +--------------------------------+

from os import system, name

def wait_input():
    input('\nPress RETURN to continue...\n')


def clear_screen():
    system('cls' if name == 'nt' else 'clear')
